# Copyright (c) 2014,2018 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc.
# OCaml implementation and API


import json
from fog05_sdk.interfaces.States import State
from fog05_sdk.types import user_fdu
from pyangbind.lib.serialise import pybindJSONDecoder
from pyangbind.lib.serialise import pybindJSONEncoder
from collections import OrderedDict


class FDU(object):

    def __init__(self, data=None):
        '''

        Constructor for the FDU Descriptor

        Parameters
        ----------
        data : dictionary
            optional dictionart containing the FDU descriptor

        returns
        -------
        FDU

        '''

        self.fdu = user_fdu.user_fdu()
        self.encoder = pybindJSONEncoder()
        self.id = None
        self.uuid = None
        self.name = None
        self.description = None
        self.image = None
        self.command = None
        self.storage = []
        self.computation_requirements = None
        self.geographical_requirements = None
        self.energy_requirements = None
        self.hypervisor = None
        self.migration_kind = None
        self.configuration = None
        self.interfaces = []
        self.io_ports = []
        self.connection_points = []
        self.depends_on = []
        if data is not None:
            pybindJSONDecoder.load_ietf_json({'fdu_descriptor':data}, None, None, obj=self.fdu, skip_unknown=True)
            self.enforce()

            self.id = self.fdu.fdu_descriptor.id
            self.uuid = data.get('uuid', None)
            self.image = data.get('image', None)
            self.command = data.get('command', None)
            self.name =  data.get('name')
            self.description =  data.get('description', None)
            self.storage = data.get('storage')
            self.computation_requirements = data.get('computation_requirements')
            self.geographical_requirements = data.get('geographical_requirements', None)
            self.energy_requirements = data.get('energy_requirements', None)
            self.hypervisor = data.get('hypervisor')
            self.migration_kind = data.get('migration_kind')
            self.configuration = data.get('configuration', None)
            self.interfaces = data.get('interfaces')
            self.io_ports = data.get('io_ports')
            self.connection_points = data.get('connection_points')
            self.depends_on = data.get('depends_on')


    def enforce(self):
        if self.fdu.fdu_descriptor.id == '':
            raise ValueError('FDU.ID cannot be empty')

        if self.fdu.fdu_descriptor.name == '':
            raise ValueError('FDU.Name cannot be empty')

        if self.fdu.fdu_descriptor.computation_requirements.cpu_arch == '':
            raise ValueError('FDU.Computation_Requirements.CPU_Arch cannot be empty')

        if self.fdu.fdu_descriptor.hypervisor == '':
            raise ValueError('FDU.Hypervisor cannot be empty')

        if self.fdu.fdu_descriptor.migration_kind == '':
            raise ValueError('FDU.Migration_Kind cannot be empty')

    def to_json(self):
        '''
        Converts the FDU into a dictionary

        returns
        -------
        dictionary
        '''
        data = {
            'uuid': self.uuid,
            'id': self.id,
            'name': self.name,
            'description': self.description,
            'image': self.image,
            'command': self.command,
            'storage': self.storage,
            'computation_requirements': self.computation_requirements,
            'geographical_requirements': self.geographical_requirements,
            'energy_requirements': self.energy_requirements,
            'hypervisor': self.hypervisor,
            'migration_kind': self.migration_kind,
            'configuration': self.configuration,
            'interfaces': self.interfaces,
            'io_ports': self.io_ports,
            'connection_points': self.connection_points,
            'depends_on': self.depends_on,
        }
        check_obj = user_fdu.user_fdu()
        pybindJSONDecoder.load_ietf_json({'fdu_descriptor':data}, None, None, obj=check_obj, skip_unknown=True)
        return data

    def get_uuid(self):
        return self.uuid

    def set_uuid(self, uuid) :
        self.uuid = uuid

    def get_id(self):
        return self.id

    def set_id(self, id) :
        self.id = id

    def get_description(self):
        return self.description

    def set_description(self, description) :
        self.description = description

    def get_name(self):
        return self.name

    def set_name(self, name) :
        self.name = name

    def get_image(self):
        return self.image

    def set_image(self, image) :
        self.image = image

    def get_image_uri(self):
        return self.image.get('uri')

    def set_image_uri(self, uri) :
        self.image.update({'uri':uri})

    def get_image_uuid(self):
        return self.image.get('uuid')

    def set_image_uuid(self, uuid) :
        self.image.update({'uuid':uuid})

    def get_image_format(self):
        return self.image.get('format')

    def set_image_format(self, format) :
        self.image.update({'format':format})

    def get_command(self):
        return self.command

    def set_command(self, command) :
        self.command = command

    def get_storage(self):
        return self.storage

    def set_storage(self, storage) :
        self.storage = storage

    def get_computation_requirements(self):
        return self.computation_requirements

    def set_computation_requirements(self, computation_requirements) :
        self.computation_requirements = computation_requirements

    def get_geographical_requirements(self):
        return self.geographical_requirements

    def set_geographical_requirements(self, geographical_requirements) :
        self.geographical_requirements = geographical_requirements

    def get_energy_requirements(self):
        return self.energy_requirements

    def set_energy_requirements(self, energy_requirements) :
        self.energy_requirements = energy_requirements

    def get_hypervisor(self):
        return self.hypervisor

    def set_hypervisor(self, hypervisor) :
        self.hypervisor = hypervisor

    def get_configuration(self):
        return self.configuration

    def set_configuration(self, configuration) :
        self.configuration = configuration

    def get_io_ports(self):
        return self.io_ports

    def set_io_ports(self, io_ports) :
        self.io_ports = io_ports

    def get_interfaces(self):
        return self.interfaces

    def set_interfaces(self, interfaces) :
        self.interfaces = interfaces

    def get_connection_points(self):
        return self.connection_points

    def set_connection_points(self, connection_points) :
        self.connection_points = connection_points

    def get_depends_on(self):
        return self.depends_on

    def set_depends_on(self, depends_on) :
        self.depends_on = depends_on

    def __str__(self):
        return "Name : {} ID: {}".format(self.name, self.id)



# Copyright (c) 2014,2018 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc.
# Initial implementation and API

import time
from fog05_sdk.interfaces.Plugin import Plugin


class NetworkPlugin(Plugin):
    '''
    Class: NetworkPlugin

    This class is an interface for plugins that control the network resouces,
    and provides an abstraction layer
    for networking managment functions
    '''

    def __init__(self, version, plugin_uuid):
        super(NetworkPlugin, self).__init__(version, plugin_uuid)
        self.netns = {}
        self.ns_index = 0

    def wait_dependencies(self):
        '''
        Waits for the Agent and the OS plugin to be ready
        '''
        self.get_agent()
        os = None
        while os is None:
            try:
                os = self.get_os_plugin()
            except (RuntimeError, ValueError):
                time.sleep(1)
        return

    def create_virtual_interface(self, intf_id, descriptor):
        '''
        This should create a virtual network interface

        paramters
        ---------
        intf_id : string
            UUID of the virtual interface
        descriptor : dictionary
            descriptor of the virtual interface

        returns
        ---------
        dictionary
            in the form of {'result':interface_info}

        '''
        raise NotImplementedError('This is and interface!')

    def create_virtual_bridge(self, name, uuid):
        '''
        This should create a virtual bridge

        parameters
        -----------
        name : string
            name of the virtual bridge to be created


        returns
        -------
        dictionary

            in the form of {'result':{'uuid': bridge uuid, 'name':bridge name}}
        '''
        raise NotImplementedError('This is and interface!')

    def allocate_bandwidth(self, intf_uuid, bandwidth):
        '''
        This should allocate bandwidth to a given virtual interface,
        if the interface not exists throw an exception

        parameters
        ----------
        intf_uuid : string

            UUID of the virtual interface

        bandwidth : int

            bandwidth to be allocated in Mbps

        returns
        -------
        bool

        '''
        raise NotImplementedError('This is and interface!')

    def create_virtual_network(self, network_name, uuid, ip_range, has_dhcp,
                               gateway, manifest):
        '''
        This should create a virtual network, with given caratteristics


        parameters
        ----------


        network_name : string

            name of the virtual network to be created

        uuid : string

            UUID of the virtual network to be created

        ip_range : string

            range should specified as CIRD subnet
            eg. 192.168.0.0/24 which means from 192.168.0.1 to 192.168.0.254

        has_dhcp : bool

            if the virtual network will have a DHCP server

        gateway : string

            IP address of the default gateway,if  None the FDUS connected to the
            network cannot reach internet

        manifest : dictionary

            the virtual network descriptor


        returns
        -------

        (string, string)

            in the form of (net_name,net_uuid)

        '''

        raise NotImplementedError('This is and interface!')

    def delete_virtual_interface(self, intf_id):
        '''
        Deletes the given virtual interface


        parameters
        ---------

        intf_id : string


        returns
        -------
        dictionary

            {'result':interface info}

        '''

        raise NotImplementedError('This is and interface!')

    def delete_virtual_bridge(self, br_uuid):
        '''
        Deletes a virtual bride, if the bridge is one assigned to a network
         should throw an exception, if the bridge not exists throw an exception

        parameters
        ----------

        br_uuid : string

            bridge UUID

        returns
        -------

        dictionary

            {'result': bridge UUID}

        '''

        raise NotImplementedError('This is and interface!')

    def delete_virtual_network(self, network_uuid):
        '''
        Delete the virtual network network_uuid, for correct network shutdown
         should kill the dnsmasq process eventually associated
        for dhcp and remove the bridge, if there are interface associate
         to this network should throw an exception


        parameters
        ----------


        network_uuid : string

            Network UUID

        returns
        -------
        bool

        '''

        raise NotImplementedError('This is and interface!')

    def get_virtual_bridges_in_node(self):
        '''
        Gets a list with the names of the current virtual bridges running in the node

        returns
        ------
        string list

        '''

    def stop_network(self):
        '''
        Stops the networking plugin
        '''
        raise NotImplementedError

    def get_network_info(self, network_uuid):
        '''
        Gives information on a given virtual network


        parameters
        ---------

        network_uuid : string

            network UUID

        returns
        -------

        dictionary

        '''
        raise NotImplementedError

    def create_bridges_if_not_exist(self, expected_bridges):
        '''
        Creates the bridges missing after checking the manifest file

        parameters
        ----------

        expected_bridges : string list

            names of expected bridges

        returns
        -------
        dictionary

            in the form {'result': string list}

        '''
        raise NotImplementedError

    def connect_interface_to_connection_point(self, intf_id, cp_id):
        '''
        Connects the given interace to the given connection point

        parameters
        ----------
        intf_id : string
            ID of the virtual interface
        cp_id : string
            UUID of the connection point

        returns
        -------
        dictionary
            {'result':{'int':string, 'ext':string}}
        '''
        raise NotImplementedError

    def disconnect_interface(self, intf_id):
        '''
        Disconnects the given interface

        parameters
        ----------
        intf_id : string
            ID of the virtual interace

        returns
        -------
        dictionary
            {'result':{'int':string, 'ext':string}}
        '''
        raise NotImplementedError

    def connect_cp_to_vnetwork(self, cp_id, vnet_id):
        '''
        Connects the given connection point to the given network

        parameters
        ----------
        cp_id : string
            UUID of the connection point
        vnet_id : string
            UUID of the virtual network

        returns
        -------
        dictionary
            {'result':{'int':string, 'ext':string}}

        '''
        raise NotImplementedError

    def disconnect_cp(self, cp_id):
        '''
        Disconnects the given connection point

        parameters
        ----------
        cp_id : string
            UUID of connection point

        returns
        -------
        dictionary
            {'result':{'int':string, 'ext':string}}
        '''
        raise NotImplementedError

    def delete_port(self, cp_id):
        '''
        Deletes the given connection point

        parameters
        ----------
        cp_id : string
            UUID of the connection point

        returns
        -------
        dictionary
            {'result':dictionary}
        '''
        raise NotImplementedError

    def get_address(self, mac_address):
        '''
        Gets the IP address associated to the interface with the given MAC address

        parameters
        ----------
        mac_address : string
            the MAC address of the interface

        returns
        -------
        dictionary
            {'result':string}
        '''
        raise NotImplementedError

    def add_port_to_router(self, router_id, port_type, vnet_id=None, ip_address=None):
        '''
        Adds a port to the given virtual router

        parameters
        ----------
        router_id : string
            UUID of the virtual router
        port_type : string
            kind of the port to be added (INTERNAL, EXTERNAL)
        vnet_id : string
            eventual network to be connected
        ip_address : string
            eventual address for the new router port

        returns
        -------
        dictionary
        '''
        raise NotImplementedError

    def remove_port_from_router(self, router_id, vnet_id):
        '''
        Removes a port from the given router

        parameters
        ----------
        router_id : string
            UUID of the virtual router
        vnet_id : string
            network to be disconnected

        returns
        -------
        dictionary
            {'result':dictionary}
        '''
        raise NotImplementedError

    def create_floating_ip(self):
        '''
        Creates a floating IP

        returns
        -------
        dictionary
            {'result':dictionary}
        '''
        raise NotImplementedError

    def delete_floating_ip(self, ip_id):
        '''
        Deletes the given floating IP

        parameters
        ----------
        ip_id : string
            UUID of the floating IP

        returns
        -------
        dictionary
            {'result':dictionary}
        '''
        raise NotImplementedError

    def assign_floating_ip(self, ip_id, cp_id):
        '''
        Assigns the given floating IP to the given conncetion point

        parameters
        ----------
        ip_id : string
            UUID of the floating IP
        cp_id : string
            UUID of the connection point

        returns
        -------
        dictionary
            {'result':dictionary}
        '''
        raise NotImplementedError

    def remove_floating_ip(self, ip_id, cp_id):
        '''
        Retains the given floating IP to the given conncetion point

        parameters
        ----------
        ip_id : string
            UUID of the floating IP
        cp_id : string
            UUID of the connection point

        returns
        -------
        dictionary
            {'result':dictionary}
        '''
        raise NotImplementedError

    def get_overlay_face(self):
        '''
        Gets the configured interface for overlay networks

        returns
        -------
        string
        '''
        raise NotImplementedError

    def get_vlan_face(self):
        '''
        Gets the configured interface for VLAN networks

        returns
        -------
        string
        '''
        raise NotImplementedError


    def create_network_namespace(self):
        '''
        Creates a new network namespace, and returns its name

        returns
        -------
        {'result':string}
        '''
        raise NotImplementedError

    def delete_network_namespace(self, nsname):
        '''
        Deletes the given network namespace, and returns its name

        returns
        -------
        {'result':string}
        '''
        raise NotImplementedError


    def create_virtual_interface_in_namespace(self, internal_name, nsname):
        '''
        Creates a veth pair in the given network namespace, with the given name for the internal interface

        returns
        -------
        {'result': dictionary }
        '''
        raise NotImplementedError

    def delete_virtual_interface_from_namespace(self, internal_name, nsname):
        '''
        Deletes the given interface from the the given network namespace

        returns
        -------
        {'result': dictionary }
        '''
        raise NotImplementedError


    def assign_address_to_interface_in_namespace(self, intf_name, nsname, address):
        '''
        Assigns the given address to the given interface in the the given network namespace

        Address are in the form AAA.AAA.AAA.AAA/NM

        returns
        -------
        {'result': dictionary }
        '''
        raise NotImplementedError

    def get_address_of_interface_in_namespace(self, intf_name, nsname):
        '''
        Retrieves the address to the given interface in the the given network namespace

        returns
        -------
        {'result': dictionary }
        '''
        raise NotImplementedError

    def remove_address_from_interface_in_namespace(self, intf_name, nsname):
        '''
        Removes the address from the given interface in the the given network namespace
        returns
        -------
        {'result': dictionary }
        '''
        raise NotImplementedError


class BridgeAssociatedToNetworkException(Exception):
    def __init__(self, message, errors=0):
        super(BridgeAssociatedToNetworkException, self).__init__(message)
        self.errors = errors


class NetworkAlreadyExistsException(Exception):
    def __init__(self, message, errors=0):
        super(NetworkAlreadyExistsException, self).__init__(message)
        self.errors = errors


class NetworkHasPendingInterfacesException(Exception):
    def __init__(self, message, errors=0):

        super(NetworkHasPendingInterfacesException, self).__init__(message)
        self.errors = errors


class InterfaceNotInNetworkException(Exception):
    def __init__(self, message, errors=0):

        super(InterfaceNotInNetworkException, self).__init__(message)
        self.errors = errors


class BridgeNotExistingException(Exception):
    def __init__(self, message, errors=0):

        super(BridgeNotExistingException, self).__init__(message)
        self.errors = errors


class InterfaceNotExistingException(Exception):
    def __init__(self, message, errors=0):

        super(InterfaceNotExistingException, self).__init__(message)
        self.errors = errors



# Copyright (c) 2014,2018 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc.
# Initial implementation and API

from fog05_sdk.interfaces.Plugin import Plugin
import uuid


class OSPlugin(Plugin):
    '''
    Class: OSPlugin

    This class is an interface for plugins that allow interaction with underlying operating
    system provides an abstraction layer for some management
    and monitoring functions

    '''

    def __init__(self, version, plugin_uuid=None):
        super(OSPlugin, self).__init__(version, plugin_uuid)
        self.requirements = []

    def get_base_path(self):
        raise NotImplementedError

    def execute_command(self, command, blocking, external):
        '''
        Executes a command on underlying os,

        parameters
        ---------

        command : string
            command to be executed

        blocking : bool
            true if the call has to block until the end of the command

        external : bool
            true if the command has to be executed in an external os shell

        returns
        ------
        dictionary

            {'result': string}

        '''

        raise NotImplementedError('This is and interface!')

    def store_file(self, content, file_path, filename):
        '''
        Stores a file in local disk, maybe can convert
         from windows dir separator to unix dir separator

        parameters
        ----------

        content : string
            base64 encoded and hexified bytes of the file content

        file_path : string
            path where the content will stored

        filename : string
            name of the file

        returns
        -------
        dictionary
            {'result':bool}
        '''

        raise NotImplementedError('This is and interface!')

    def checksum(self, file_path):
        '''
        Calculates the SHA256 checksum for the given file

        parameters
        -----------
        file_path : string
            path to file

        returns
        -------
        dictionary
            {'result':string}
        '''
        raise NotImplementedError('This is and interface!')
    def file_exists(self, file_path):
        '''
        Checks if the given file exists

        parameters
        ----------
        file_path : string
            path to the file

        returns
        -------
        dictionary
            {'result':bool}
        '''
        raise NotImplementedError

    def dir_exists(self, dir_path):
        '''
        Checks if the given directory exists

        parameters
        ----------
        dir_path : string
            path to the directory

        returns
        -------
        dictionary
            {'result':bool}


        '''
        raise NotImplementedError

    def create_dir(self, dir_path):
        '''
        Creates the given new directory

        parameters
        ----------
        dir_path : string
            path to the directory

        returns
        -------
        dictionary
            {'result':bool}

        '''
        raise NotImplementedError

    def create_file(self, file_path):
        '''
        Creates the given new empty file

        parameters
        ----------
        file_path : string
            path to the file

        returns
        -------
        dictionary
            {'result':bool}
        '''
        raise NotImplementedError

    def remove_dir(self, dir_path):
        '''
        Removes the given directory

        parameters
        ----------
        dir_path : string
            dir_path to the directory

        returns
        -------
        dictionary
            {'result':bool}

        '''
        raise NotImplementedError

    def remove_file(self, file_path):
        '''
        Removes the given file

        parameters
        ----------
        file_path : string
            path to the directory

        returns
        -------
        dictionary
            {'result':bool}
        '''
        raise NotImplementedError

    def read_file(self, file_path, root=False):
        '''
        Read the content from a file in the local disk,
         maybe can convert from windows dir separator to unix dir separator
        return the file content
        throw an exception if file not exits

        parameters
        ----------

        file_path : string
            path to file

        root : bool
            if true it will use sudo cat to read the file

        returns
        --------
        dictionary
            {'result':bytes}


        '''

        raise NotImplementedError('This is and interface!')

    def get_neighbors(self):
        '''
        Gets LLDP neighbors

        retutns
        ------
        list of dictionaries

        '''
        raise NotImplementedError('This is and interface!')

    def read_binary_file(self, file_path):
        '''
        Reads binary file

        parameters
        ----------
        file_path : string
            path to file

        returns
        -------
        bytes

        '''
        raise NotImplementedError

    def download_file(self, url, file_path):
        '''
        Downloads the given file in the given path

        parameters
        ----------
        url : string
            url for the source file
        file_path : string
            path to destination file

        returns
        ------
        dictionary
            {'result':bool}
        '''
        raise NotImplementedError

    def get_CPU_level(self):
        '''
        Gets the current CPU usage

        returns
        -------
        float
        '''

        raise NotImplementedError('This is and interface!')

    def get_memory_level(self):
        '''
        Gets the current RAM usage

        returns
        -------
        float
        '''
        raise NotImplementedError('This is and interface!')

    def get_storage_level(self):
        '''
        Get the current root disk usage

        returns
        -------
        float
        '''
        raise NotImplementedError('This is and interface!')

    def get_network_level(self):
        '''
        Gets the current network usage statistics

        returns
        -------
        list of dictionaries
        '''
        raise NotImplementedError('This is and interface!')

    def install_package(self, packages):
        '''
        Installs all packages passed within the parameter, return a bool
        to know the retult of operation

        parameters
        ----------

        packages : string list

            name of the packages to be installed

        returns
        -------
        list (string, bool)

            name of package, bool
        '''

        raise NotImplementedError('This is and interface!')

    def remove_package(self, packages):
        '''
        Removes the given packages

        parameters
        ---------
        packages : list of string
            packages to be removed

        returns
        -------
        list (string, bool)
            name of package, bool
        '''

        raise NotImplementedError('This is and interface!')

    def check_if_pid_exists(self, pid):
        '''
        Checks if a  given PID exists

        parameters
        ----------
        pid : int
            PID to be verified

        returns
        -------
        dictionary
            {'result':bool}
        '''
        raise NotImplementedError('This is and interface!')

    def send_signal(self, signal, pid):
        '''
        Sends a signal to the given pid

        parametes
        ---------
        signal : int
            signal to be sent
        pid : int
            PID to be signaled

        returns
        -------
        dictionary
            {'result':bool}

        '''

        raise NotImplementedError('This is and interface!')

    def send_sig_int(self, pid):
        '''
        Sends a SigKill (Ctrl+C) to the given PID

        parameters
        ----------
        pid : int
            pid to be signaled

        returns
        -------
        dictionary
            {'result':bool}
        '''

        raise NotImplementedError('This is and interface!')

    def send_sig_kill(self, pid):
        '''
        Sends a SigKill (kill the process) to the given pid
        throw an exception if pid not exits

        parameters
        ----------
        pid : int
            pid to be signaled

        returns
        -------
        dictionary
            {'result':bool}
        '''

        raise NotImplementedError('This is and interface!')

    def get_uuid(self):
        '''
        Gets the node UUID

        returns
        -------
        string
        '''
        raise NotImplementedError('This is and interface!')

    def get_processor_information(self):
        '''
        Gets information on the node CPU

        returns
        -------
        list of dictionaries
            {'model':string, 'frequency':float, 'arch':string}
        '''
        raise NotImplementedError

    def get_memory_information(self):
        '''
        Gets information on the node RAM

        returns
        -------
        dictionary
            {'size':float}

        '''
        raise NotImplementedError

    def get_disks_information(self):
        '''
        Gets information on the node disks usage

        returns
        -------
        list of dictionaries
            {'local_address':string, 'dimension':float, 'mount_point':string, 'filesystem':string}
        '''
        raise NotImplementedError

    def get_io_informations(self):
        '''
        Gets information about node IO ports

        returns
        -------
        list of dictionaries
            {'name':string, 'io_type':string, 'io_file':string, 'available':bool}
        '''
        raise NotImplementedError

    def get_accelerators_informations(self):
        '''
        Gets information about node hardware accelerators

        returns
        -------
        list of dictionaries
            {'hw_address':string, 'name':string, 'supported_library' string list, 'available':bool}

        '''
        raise NotImplementedError

    def get_network_informations(self):
        '''
        Gets information about node network interfaces

        returns
        -------
        list of dictionaties
            {
                'intf_name':string,
                'intf_mac_address':string,
                'intf_speed': int,
                'type':string,
                'available':bool,
                'default_gw':bool,
                'intf_configuration':
                {
                    'ipv4_address':string,
                    'ipv4_netmask':string,
                    'ipv4_gateway':string.
                    'ipv6_address':string,
                    'ipv6_netmask':string,
                    'ipv6_gateway':string.
                    'bus_address':string
                }
            }

        '''
        raise NotImplementedError

    def get_position_information(self):
        '''
        Gets information about mode position

        returns
        -------
        dictionary
            {'lat':float, 'lon':float}
        '''
        raise NotImplementedError

    def add_know_host(self, hostname, ip):
        '''
        Adds the given host in node ssh configuration

        parameters
        ----------
        hostname : string
            host to be added
        ip : string
            IP address of the host

        returns
        -------
        dictionary
            {'result':bool}
        '''
        raise NotImplementedError

    def remove_know_host(self, hostname):
        '''
        Removes the given host from node ssh configuration

        parameters
        ----------
        hostname : string
            host to be removed

        returns
        -------
        dictionary
            {'result':bool}
        '''
        raise NotImplementedError

    def get_intf_type(self, name):
        '''
        Gets the inteface type for the given interface

        parameters
        ----------
        name : string
            name of the interface

        returns
        -------
        dictionary
            {'result':string}
        '''
        raise NotImplementedError

    def set_io_unaviable(self, io_name):
        '''
        Sets a given IO device as unavailable

        paramters
        ---------
        io_name : string
            name of the IO device

        returns
        -------
        bool
        '''
        raise NotImplementedError

    def set_io_available(self, io_name):
        '''
        Sets a given IO device as available

        paramters
        ---------
        io_name : string
            name of the IO device

        returns
        -------
        bool
        '''
        raise NotImplementedError

    def set_accelerator_unaviable(self, acc_name):
        '''
        Sets a given accelerator device as unavailable

        paramters
        ---------
        acc_name : string
            name of the accelerator device

        returns
        -------
        bool
        '''
        raise NotImplementedError

    def set_accelerator_available(self, acc_name):
        '''
        Sets a given accelerator device as available

        paramters
        ---------
        acc_name : string
            name of the accelerator device

        returns
        -------
        bool
        '''
        raise NotImplementedError

    def set_interface_unaviable(self, intf_name):
        '''
        Sets a given network device as unavailable

        paramters
        ---------
        intf_name : string
            name of the network device

        returns
        -------
        bool
        '''
        raise NotImplementedError

    def set_interface_available(self, intf_name):
        '''
        Sets a given network device as available

        paramters
        ---------
        intf_name : string
            name of the network device

        returns
        -------
        bool
        '''
        raise NotImplementedError

    def get_hostname(self):
        '''
        Gets node hostname

        returns
        -------
        string
        '''
        raise NotImplementedError

    def local_mgmt_address(self):
        '''
        Gets node management IP address

        returns
        -------
        dictionary
            {'result':string}
        '''
        raise NotImplementedError


class ProcessNotExistingException(Exception):
    def __init__(self, message, errors=0):

        super(ProcessNotExistingException, self).__init__(message)
        self.errors = errors


class FileNotExistingException(Exception):
    def __init__(self, message, errors=0):

        super(FileNotExistingException, self).__init__(message)
        self.errors = errors



# Copyright (c) 2014,2019 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc.
# Connector and helper for YAKS


import json
import concurrent.futures
from threading import Thread
from fog05_sdk.interfaces import Constants
from yaks import Yaks, Value, Encoding, Change



class GAD(object):
    def __init__(self, workspace, prefix):
        self.ws = workspace
        self.prefix = prefix
        self.listeners = []
        self.evals = []

    def unsubscribe(self, subid):
        if subid in self.listeners:
            self.ws.unsubscribe(subid)
            self.listeners.remove(subid)

    def unregister_eval(self, path):
        if path in self.evals:
            self.ws.unregister_eval(path)
            self.evals.remove(path)

    def close(self):
        for s in self.listeners:
            self.ws.unsubscribe(s)
        for ep in self.evals:
            self.ws.unregister_eval(ep)

    # SYSTEM

    def get_sys_info_path(self, sysid):
        return Constants.create_path([self.prefix, sysid, 'info'])

    def get_sys_configuration_path(self, sysid):
        return Constants.create_path([self.prefix, sysid, 'configuration'])

    def get_all_users_selector(self, sysid):
        return Constants.create_path([self.prefix, sysid, 'users', '*'])

    def get_user_info_path(self, sysid, userid):
        return Constants.create_path(
            [self.prefix, sysid, 'users', userid, 'info'])

    # TENANTS

    def get_all_tenants_selector(self, sysid):
        return Constants.create_path([self.prefix, sysid, 'tenants', '*'])

    def get_tenant_info_path(self, sysid, tenantid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants', tenantid, 'info'])

    def get_tenant_configuration_path(self, sysid, tenantid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants', tenantid, 'configuration'])

    # CATALOG

    def get_catalog_atomic_entity_info_path(self, sysid, tenantid, aeid):
        return Constants.create_path([self.prefix, sysid, 'tenants',
            tenantid, 'catalog', 'atomic-entities', aeid, 'info'])

    def get_catalog_all_atomic_entities_selector(self, sysid, tenantid):
        return Constants.create_path([self.prefix, sysid, 'tenants',
            tenantid, 'catalog', 'atomic-entities', '*', 'info'])

    def get_catalog_fdu_info_path(self, sysid, tenantid, fduid):
        return Constants.create_path([self.prefix, sysid, 'tenants',
            tenantid, 'catalog', 'fdu', fduid, 'info'])

    def get_catalog_all_fdu_selector(self, sysid, tenantid):
        return Constants.create_path([self.prefix, sysid, 'tenants',
            tenantid, 'catalog', 'fdu', '*', 'info'])

    def get_catalog_entity_info_path(self, sysid, tenantid, aeid):
        return Constants.create_path([self.prefix, sysid, 'tenants',
            tenantid, 'catalog', 'entities', aeid, 'info'])

    def get_catalog_all_entities_selector(self, sysid, tenantid):
        return Constants.create_path([self.prefix, sysid, 'tenants',
            tenantid, 'catalog', 'entities', '*', 'info'])

    # RECORDS

    def get_records_atomic_entity_instance_info_path(self, sysid, tenantid, aeid, instanceid):
        return Constants.create_path([self.prefix, sysid, 'tenants',
            tenantid, 'records', 'atomic-entities', aeid, 'instances', instanceid, 'info'])

    def get_records_all_atomic_entity_instances_selector(self, sysid, tenantid, aeid):
        return Constants.create_path([self.prefix, sysid, 'tenants',
            tenantid, 'records', 'atomic-entities', aeid, 'instances','*', 'info'])

    def get_records_all_atomic_entities_instances_selector(self, sysid, tenantid):
        return Constants.create_path([self.prefix, sysid, 'tenants',
            tenantid, 'catalog', 'atomic-entities', '*', 'instances', '*' ,'info'])

    def get_records_entity_instance_info_path(self, sysid, tenantid, aeid, instanceid):
        return Constants.create_path([self.prefix, sysid, 'tenants',
            tenantid, 'records', 'entities', aeid, 'instances', instanceid, 'info'])

    def get_records_all_entity_instances_selector(self, sysid, tenantid, aeid):
        return Constants.create_path([self.prefix, sysid, 'tenants',
            tenantid, 'records', 'entities', aeid, 'instances','*', 'info'])

    def get_records_all_entities_instances_selector(self, sysid, tenantid):
        return Constants.create_path([self.prefix, sysid, 'tenants',
            tenantid, 'catalog', 'entities', '*', 'instances', '*' ,'info'])

    # Nodes

    def get_all_nodes_selector(self, sysid, tenantid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants', tenantid, 'nodes', '*', 'info'])

    def get_node_info_path(self, sysid, tenantid, nodeid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'info'])

    def get_node_configuration_path(self, sysid, tenantid, nodeid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'configuration'])

    def get_node_status_path(self, sysid, tenantid, nodeid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'status'])

    def get_node_plugins_selector(self, sysid, tenantid, nodeid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'plugins', '*', 'info'])

    def get_node_plugins_subscriber_selector(self, sysid, tenantid, nodeid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'plugins', '**'])

    def get_node_plugin_info_path(self, sysid, tenantid, nodeid, pluginid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'plugins',
                                      pluginid, 'info'])

    def get_node_plugin_eval_path(self, sysid, tenantid, nodeid, pluginid,
                                  func_name):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'plugins',
                                      pluginid, 'exec', func_name])

    # Node FDU or Records

    def get_node_fdu_info_path(self, sysid, tenantid, nodeid, fduid, instanceid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'fdu',fduid ,
                                      'instances', instanceid, 'info'])

    def get_node_fdu_selector(self, sysid, tenantid, nodeid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'fdu', '*', 'instances',
                                      '*', 'info'])

    def get_node_fdu_instances_selector(self, sysid, tenantid, nodeid, fduid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'fdu', fduid, 'instances',
                                      '*', 'info'])

    def get_node_fdu_instance_selector(self, sysid, tenantid, nodeid, instid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'fdu', '*', 'instances',
                                      instid, 'info'])

    def get_fdu_instance_selector(self, sysid, tenantid, instid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', '*', 'fdu', '*', 'instances',
                                      instid, 'info'])

    def get_fdu_start_instance_selector(self, sysid, tenantid, instid, env):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', '*', 'fdu', '*', 'instances',
                                      instid, 'start', '?(env={})'.format(env)])

    def get_fdu_run_instance_selector(self, sysid, tenantid, instid, env):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', '*', 'fdu', '*', 'instances',
                                      instid, 'run', '?(env={})'.format(env)])

    def get_fdu_log_instance_selector(self, sysid, tenantid, instid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', '*', 'fdu', '*', 'instances',
                                      instid, 'log'])
    def get_fdu_ls_instance_selector(self, sysid, tenantid, instid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', '*', 'fdu', '*', 'instances',
                                      instid, 'ls'])
    def get_fdu_file_instance_selector(self, sysid, tenantid, instid, filename):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', '*', 'fdu', '*', 'instances',
                                      instid, 'get', '?(filename={})'.format(filename)])

    def get_fdu_start_instance_path(self, sysid, tenantid,nodeid, fduid, instid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'fdu', fduid, 'instances',
                                      instid, 'start'])

    def get_fdu_run_instance_path(self, sysid, tenantid,nodeid, fduid, instid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'fdu', fduid, 'instances',
                                      instid, 'run'])

    def get_fdu_log_instance_path(self, sysid, tenantid,nodeid, fduid, instid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'fdu', fduid, 'instances',
                                      instid, 'log'])
    def get_fdu_ls_instance_path(self, sysid, tenantid,nodeid, fduid, instid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'fdu', fduid, 'instances',
                                      instid, 'ls'])
    def get_fdu_file_instance_path(self, sysid, tenantid,nodeid, fduid, instid):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'fdu', fduid, 'instances',
                                      instid, 'get'])

    # Scheduler

    def get_fdu_scheduler_selector(self, sysid, tenantid, nodeid, fdu):
        return Constants.create_path([self.prefix, sysid, 'tenants', tenantid,
                                      'nodes', nodeid, 'agent', 'exec','schedule',
                                    '?(fdu_id={})'.format(fdu)])

    # Networks

    def get_all_networks_selector(self, sysid, tenantid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants', tenantid, 'networks', '*', 'info'])

    def get_entity_all_instances_selector(self, sysid, tenantid, entityid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants', tenantid,
             'entities',  entityid, 'instances', '*'])

    def get_entity_info_path(self, sysid, tenantid, entityid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants',
             tenantid, 'entities', entityid, 'info'])

    def get_network_info_path(self, sysid, tenantid, networkid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants',
             tenantid, 'networks', networkid, 'info'])

    def get_network_port_info_path(self, sysid, tenantid, portid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants',
             tenantid, 'networks', 'ports', portid, 'info'])

    def get_all_ports_selector(self, sysid, tenantid):
        return Constants.create_path([
            self.prefix,sysid ,'tenants', tenantid,'networks','ports',
            '*', 'info'])

    def get_network_router_info_path(self, sysid, tenantid, routerid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants',
             tenantid, 'networks', 'routers', routerid, 'info'])

    def get_all_routers_selector(self, sysid, tenantid):
        return Constants.create_path([
            self.prefix,sysid ,'tenants', tenantid,'networks','routers',
            '*', 'info'])

    # Images

    def get_image_info_path(self, sysid, tenantid, imageid):
        return Constants.create_path([
            self.prefix, sysid, 'tenants', tenantid, 'image', imageid, 'info'
        ])

    def get_all_image_selector(self, sysid, tenantid):
        return Constants.create_path([
            self.prefix, sysid, 'tenants', tenantid, 'image', '*', 'info'
        ])

    # Node Images

    def get_node_image_info_path(self, sysid, tenantid, nodeid, imageid):
        return Constants.create_path([
            self.prefix, sysid, 'tenants', tenantid, 'nodes', nodeid,
             'image', imageid, 'info'
        ])

    def get_all_node_image_selector(self, sysid, tenantid, nodeid):
        return Constants.create_path([
            self.prefix, sysid, 'tenants', tenantid, 'nodes', nodeid,
            'image', '*', 'info'
        ])

    # Flavor

    def get_flavor_info_path(self, sysid, tenantid, flavorid):
        return Constants.create_path([
            self.prefix, sysid, 'tenants', tenantid, 'flavor', flavorid, 'info'
        ])

    def get_all_flavor_selector(self, sysid, tenantid):
        return Constants.create_path([
            self.prefix, sysid, 'tenants', tenantid, 'flavor', '*', 'info'
        ])

    # Node Flavor

    def get_node_flavor_info_path(self, sysid, tenantid, nodeid, flavorid):
        return Constants.create_path([
            self.prefix, sysid, 'tenants', tenantid, 'nodes', nodeid,
             'flavor', flavorid, 'info'
        ])

    def get_all_node_flavor_selector(self, sysid, tenantid, nodeid):
        return Constants.create_path([
            self.prefix, sysid, 'tenants', tenantid, 'nodes', nodeid,
            'flavor', '*', 'info'
        ])

    # Node Network

    def get_node_network_floating_ip_info_path(self, sysid, tenantid, nodeid, ipid):
         return Constants.create_path([
            self.prefix, sysid, 'tenants', tenantid, 'nodes', nodeid,
            'networks', 'floating-ips', ipid, 'info'])

    def get_node_all_network_floating_ips_selector(self, sysid, tenantid, nodeid):
         return Constants.create_path([
            self.prefix, sysid, 'tenants', tenantid, 'nodes', nodeid,
            'networks', 'floating-ips', '*', 'info'])

    def get_node_network_ports_selector(self, sysid, tenantid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants', tenantid,
            'nodes', '*', 'networks', 'ports', '*', 'info'])

    def get_node_network_port_info_path(self, sysid, tenantid, nodeid, portid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants', tenantid,
            'nodes', nodeid, 'networks', 'ports', portid, 'info'])

    def get_node_network_routers_selector(self, sysid, tenantid, nodeid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants', tenantid,
            'nodes', nodeid, 'networks', 'routers', '*', 'info'])

    def get_node_network_router_info_path(self, sysid, tenantid, nodeid, routerid):
        return Constants.create_path(
            [self.prefix, sysid, 'tenants', tenantid,
            'nodes', nodeid, 'networks', 'routers', routerid, 'info'])

    def get_node_network_info_path(self, sysid, tenantid, nodeid, networkid):
        return Constants.create_path([
            self.prefix,sysid ,'tenants', tenantid,
            'nodes', nodeid, 'networks', networkid, 'info'])


    # Evals

    def dict2args(self, d):
        i = 0
        b = ''
        for k in d:
            v = d.get(k)
            if isinstance(v,(dict, list)):
                v = json.dumps(v)
            if i == 0:
                b = b + '{}={}'.format(k, v)
            else:
                b = b + ';{}={}'.format(k, v)
            i = i + 1
        return '('+b+')'

    def get_agent_exec_path(self, sysid, tenantid, nodeid, func_name):
        return Constants.create_path([self.prefix, sysid, 'tenants',
        tenantid, 'nodes', nodeid, 'agent', 'exec', func_name])

    def get_agent_exec_path_with_params(self, sysid, tenantid, nodeid, func_name, params):
        if len(params) > 0:
            p = self.dict2args(params)
            f = func_name + '?' + p
        else:
            f = func_name
        return Constants.create_path([self.prefix, sysid, 'tenants',
        tenantid, 'nodes', nodeid, 'agent', 'exec',f])

    # ID Extraction

    def extract_userid_from_path(self, path):
        return path.split('/')[4]

    def extract_tenantid_from_path(self, path):
        return path.split('/')[4]

    def extract_entityid_from_path(self, path):
        return path.split('/')[7]

    def extract_aeid_from_path(self, path):
        return path.split('/')[7]

    def extract_atomic_entity_instanceid_from_path(self, path):
        return path.split('/')[9]

    def extract_entity_id_from_path(self, path):
        return path.split('/')[7]

    def extract_entity_instanceid_from_path(self, path):
        return path.split('/')[9]

    def extract_fduid_from_path(self, path):
        return path.split('/')[7]

    def extract_nodeid_from_path(self, path):
        return path.split('/')[6]

    def extract_plugin_from_path(self, path):
        return path.split('/')[8]

    def extract_node_fduid_from_path(self, path):
        return path.split('/')[8]

    def extract_node_instanceid_from_path(self, path):
        return path.split('/')[10]

    def extract_node_port_id_from_path(self, path):
        return path.split('/')[9]

    def extract_node_router_id_from_path(self, path):
        return path.split('/')[9]

    def extract_node_floatingid_from_path(self, path):
        return path.split('/')[9]

    # System

    def get_sys_info(self, sysid):
        s = self.get_sys_info_path(sysid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty message list on get_sys_info')
        else:
            v = res[0].get_value()
            return json.loads(v.get_value())

    def get_sys_config(self, sysid):
        s = self.get_sys_configuration_path(sysid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty message list on get_sys_config')
        else:
            v = res[0].get_value()
            return json.loads(v.get_value())

    def get_all_users_ids(self, sysid):
        s = self.get_all_users_selector(sysid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        xs = map(lambda x: self.extract_userid_from_path(x.get_path()), res)
        return list(xs)

    # Tenants

    def get_all_tenants_ids(self, sysid):
        s = self.get_all_tenants_selector(sysid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        xs = map(lambda x: self.extract_tenantid_from_path(x.get_path()), res)
        return list(xs)

    def get_all_nodes(self, sysid, tenantid):
        s = self.get_all_nodes_selector(sysid, tenantid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        xs = map(lambda x: self.extract_nodeid_from_path(x.get_path()), res)
        return list(xs)

    def get_node_info(self, sysid, tenantid, nodeid):
        s = self.get_node_info_path(sysid, tenantid, nodeid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty message list on get_node_info')
        v = res[0].get_value()
        return json.loads(v.get_value())

    def add_node_info(self, sysid, tenantid, nodeid, nodeinfo):
        p = self.get_node_info_path(sysid, tenantid, nodeid)
        v = Value(json.dumps(nodeinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_node_info(self, sysid, tenantid, nodeid):
        p = self.get_node_info_path(sysid, tenantid, nodeid)
        self.ws.remove(p)

    def get_node_status(self, sysid, tenantid, nodeid):
        s = self.get_node_status_path(sysid, tenantid, nodeid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty message list on get_node_status')
        v = res[0].get_value()
        return json.loads(v.get_value())

    def add_node_status(self, sysid, tenantid, nodeid, nodestatus):
        p = self.get_node_status_path(sysid, tenantid, nodeid)
        v = Value(json.dumps(nodestatus), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def observe_node_status(self, sysid, tenantid, nodeid, callback):
        s = self.get_node_status_path(sysid, tenantid, nodeid)
        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def remove_node_status(self, sysid, tenantid, nodeid):
        p = self.get_node_status_path(sysid, tenantid, nodeid)
        self.ws.remove(p)

    def add_node_configuration(self, sysid, tenantid, nodeid, nodeconf):
        p = self.get_node_configuration_path(sysid, tenantid, nodeid)
        v = Value(json.dumps(nodeconf), encoding=Encoding.STRING)
        return self.ws.put(p, v)


    # Catalog

    def get_catalog_all_entities(self, sysid, tenantid):
        s = self.get_catalog_all_entities_selector(sysid, tenantid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        else:
            xs = map(lambda x: self.extract_entity_id_from_path(x.get_path()), res)
            return list(xs)

    def get_catalog_entity_info(self, sysid, tenantid, eid):
        p = self.get_catalog_entity_info_path(sysid, tenantid, eid)
        res = self.ws.get(p)
        if len(res) == 0:
            return None
        v = res[0].get_value()
        return json.loads(v.get_value())

    def add_catalog_entity_info(self, sysid, tenantid, eid, einfo):
        p = self.get_catalog_entity_info_path(sysid, tenantid, eid)
        v = Value(json.dumps(einfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_catalog_entity_info(self, sysid, tenantid, eid):
        p = self.get_catalog_entity_info_path(sysid, tenantid, eid)
        self.ws.remove(p)

    def observe_catalog_entities(self, sysid, tenantid, callback):
        s = self.get_catalog_all_entities_selector(sysid, tenantid)
        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def get_catalog_all_atomic_entities(self, sysid, tenantid):
        s = self.get_catalog_all_atomic_entities_selector(sysid, tenantid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        else:
            xs = map(lambda x: self.extract_aeid_from_path(x.get_path()), res)
            return list(xs)

    def get_catalog_atomic_entity_info(self, sysid, tenantid, aeid):
        p = self.get_catalog_atomic_entity_info_path(sysid, tenantid, aeid)
        res = self.ws.get(p)
        if len(res) == 0:
            return None
        v = res[0].get_value()
        return json.loads(v.get_value())

    def add_catalog_atomic_entity_info(self, sysid, tenantid, aeid, aeinfo):
        p = self.get_catalog_atomic_entity_info_path(sysid, tenantid, aeid)
        v = Value(json.dumps(aeinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_catalog_atomic_entity_info(self, sysid, tenantid, aeid):
        p = self.get_catalog_atomic_entity_info_path(sysid, tenantid, aeid)
        self.ws.remove(p)

    def observe_catalog_atomic_entities(self, sysid, tenantid, callback):
        s = self.get_catalog_all_atomic_entities_selector(sysid, tenantid)
        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def observe_catalog_fdu(self, sysid, tenantid, fduid,callback):
        s = self.get_catalog_fdu_info_path(sysid, tenantid, fduid)
        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid


    def get_catalog_all_fdus(self, sysid, tenantid):
        s = self.get_catalog_all_fdu_selector(sysid, tenantid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        else:
            xs = map(lambda x: self.extract_fduid_from_path(x.get_path()), res)
            return list(xs)

    def get_catalog_fdu_info(self, sysid, tenantid, fduid):
        p = self.get_catalog_fdu_info_path(sysid, tenantid, fduid)
        res = self.ws.get(p)
        if len(res) == 0:
            return None
        v = res[0].get_value()
        return json.loads(v.get_value())

    def add_catalog_fdu_info(self, sysid, tenantid, fduid, fduinfo):
        p = self.get_catalog_fdu_info_path(sysid, tenantid, fduid)
        v = Value(json.dumps(fduinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_catalog_fdu_info(self, sysid, tenantid, fduid):
        p = self.get_catalog_fdu_info_path(sysid, tenantid, fduid)
        self.ws.remove(p)

    def observe_catalog_fdus(self, sysid, tenantid, callback):
        s = self.get_catalog_all_fdu_selector(sysid, tenantid)
        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    # Records


    def get_records_all_entities_instances(self, sysid, tenantid):
        s = self.get_records_all_entities_instances_selector(sysid, tenantid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        else:
            xs = map(lambda x: (self.extract_entity_id_from_path(x.get_path()),self.extract_entity_instanceid_from_path(x.get_path())), res)
            return list(xs)

    def get_records_all_entity_instances(self, sysid, tenantid, eid):
        s = self.get_records_all_entity_instances_selector(sysid, tenantid, eid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        else:
            xs = map(lambda x: self.extract_entity_instanceid_from_path(x.get_path()), res)
            return list(xs)

    def get_records_entity_info(self, sysid, tenantid, eid, instanceid):
        p = self.get_records_entity_instance_info_path(sysid, tenantid, eid, instanceid)
        res = self.ws.get(p)
        if len(res) == 0:
            return None
        v = res[0].get_value()
        return json.loads(v.get_value())

    def add_records_entity_info(self, sysid, tenantid, eid, instanceid, aeinfo):
        p = self.get_records_entity_instance_info_path(sysid, tenantid, eid, instanceid)
        v = Value(json.dumps(aeinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_records_entity_info(self, sysid, tenantid, eid, instanceid):
        p = self.get_records_entity_instance_info_path(sysid, tenantid, eid, instanceid)
        self.ws.remove(p)

    def observe_records_entities(self, sysid, tenantid, callback):
        s = self.get_records_all_entities_instances_selector(sysid, tenantid)
        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def get_records_all_atomic_entities_instances(self, sysid, tenantid):
        s = self.get_records_all_atomic_entities_instances_selector(sysid, tenantid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        else:
            xs = map(lambda x: (self.extract_aeid_from_path(x.get_path()),self.extract_atomic_entity_instanceid_from_path(x.get_path())), res)
            return list(xs)

    def get_records_all_atomic_entity_instances(self, sysid, tenantid, aeid):
        s = self.get_records_all_atomic_entity_instances_selector(sysid, tenantid, aeid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        else:
            xs = map(lambda x: self.extract_atomic_entity_instanceid_from_path(x.get_path()), res)
            return list(xs)

    def get_records_atomic_entity_info(self, sysid, tenantid, aeid, instanceid):
        p = self.get_records_atomic_entity_instance_info_path(sysid, tenantid, aeid, instanceid)
        res = self.ws.get(p)
        if len(res) == 0:
            return None
        v = res[0].get_value()
        return json.loads(v.get_value())

    def add_records_atomic_entity_info(self, sysid, tenantid, aeid, instanceid, aeinfo):
        p = self.get_records_atomic_entity_instance_info_path(sysid, tenantid, aeid, instanceid)
        v = Value(json.dumps(aeinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_records_atomic_entity_info(self, sysid, tenantid, aeid, instanceid):
        p = self.get_records_atomic_entity_instance_info_path(sysid, tenantid, aeid, instanceid)
        self.ws.remove(p)

    def observe_records_atomic_entities(self, sysid, tenantid, callback):
        s = self.get_records_all_atomic_entities_instances_selector(sysid, tenantid)
        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid


    # Node FDU

    def add_node_fdu(self, sysid, tenantid, nodeid, fduid, instanceid ,fduinfo):
        p = self.get_node_fdu_info_path(sysid, tenantid, nodeid, fduid, instanceid)
        v = Value(json.dumps(fduinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def observe_node_fdu(self, sysid, tenantid, nodeid, fduid, instanceid, callback):
        s = self.get_node_fdu_info_path(sysid, tenantid, nodeid, fduid, instanceid)

        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def get_node_fdus(self, sysid, tenantid, nodeid):
        s = self.get_node_fdu_selector(sysid, tenantid, nodeid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        else:
            xs = map(lambda x: self.extract_node_fduid_from_path(x.get_path()), res)
            return list(xs)

    def get_node_fdu_instances(self, sysid, tenantid, nodeid, fduid):
        p = self.get_node_fdu_info_path(sysid, tenantid, nodeid, fduid, '*')
        kvs = self.ws.get(p)
        if len(kvs) == 0:
            return []
        xs = map(lambda x: (self.extract_nodeid_from_path(x.get_path()),
                    self.extract_node_fduid_from_path(x.get_path()),
                    self.extract_node_instanceid_from_path(x.get_path()),
                    json.loads(kvs[0].get_value().get_value())), kvs)
        return list(xs)

    def get_node_fdu_instance(self, sysid, tenantid, nodeid, instanceid):
        p = self.get_node_fdu_info_path(sysid, tenantid, nodeid, '*', instanceid)
        kvs = self.ws.get(p)
        if len(kvs) == 0:
            return None
        return json.loads(kvs[0].get_value().get_value())

    def get_fdu_instance_node(self, sysid, tenantid, instanceid):
        p = self.get_node_fdu_info_path(sysid, tenantid, '*', '*', instanceid)
        kvs = self.ws.get(p)
        if len(kvs) == 0:
            return None
        return self.extract_nodeid_from_path(kvs[0].get_path())

    def get_node_fdu(self, sysid, tenantid, nodeid, fduid, instanceid):
        p = self.get_node_fdu_info_path(sysid, tenantid, nodeid, fduid, instanceid)
        kvs = self.ws.get(p)
        if len(kvs) == 0:
            return None
        return json.loads(kvs[0].get_value().get_value())

    def remove_node_fdu(self, sysid, tenantid, nodeid, fduid, instanceid):
        p = self.get_node_fdu_info_path(sysid, tenantid, nodeid, fduid, instanceid)
        return self.ws.remove(p)

    def get_fdu_nodes(self, sysid, tenantid, fduid):
        s = self.get_node_fdu_info_path(sysid, tenantid, '*', fduid, '*')
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        xs = map(lambda x: self.extract_nodeid_from_path(x.get_path()), res)
        return list(xs)

    # Plugins

    def get_all_plugins_ids(self, sysid, tenantid, nodeid):
        s = self.get_node_plugins_selector(sysid, tenantid, nodeid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        xs = map(lambda x: self.extract_plugin_from_path(x.get_path()), res)
        return list(xs)

    def get_plugin_info(self, sysid, tenantid, nodeid, pluginid):
        s = self.get_node_plugin_info_path(sysid, tenantid, nodeid, pluginid)
        res = self.ws.get(s)
        if len(res) == 0:
            return None
        v = res[0].get_value()
        return json.loads(v.get_value())

    def add_node_plugin(self, sysid, tenantid, nodeid, pluginid, plugininfo):
        p = self.get_node_plugin_info_path(sysid, tenantid, nodeid, pluginid)
        v = Value(json.dumps(plugininfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def observe_node_plugins(self, sysid, tenantid, nodeid, callback):
        s = self.get_node_plugins_subscriber_selector(sysid, tenantid, nodeid)

        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def add_plugin_eval(self, sysid, tenantid, nodeid, pluginid,
                        func_name, func):
        p = self.get_node_plugin_eval_path(
            sysid, tenantid, nodeid, pluginid, func_name)

        def cb(path, properties):
            v = Value(json.dumps(func(**properties)), encoding=Encoding.STRING)
            return v
        r = self.ws.register_eval(p, cb)
        self.evals.append(p)
        return r

    # Network

    def get_network_port(self, sysid, tenantid, portid):
        s = self.get_network_port_info_path(sysid, tenantid, portid)
        kvs = self.ws.get(s)
        if len(kvs) == 0:
            return None
        return json.loads(kvs[0].get_value().get_value())

    def add_network_port(self, sysid, tenantid, portid, portinfo):
        p = self.get_network_port_info_path(sysid, tenantid, portid)
        v = Value(json.dumps(portinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_network_port(self, sysid, tenantid, portid):
        p = self.get_network_port_info_path(sysid, tenantid, portid)
        return self.ws.remove(p)

    def get_all_network_ports(self, sysid, tenantid):
        p = self.get_all_ports_selector(sysid, tenantid)
        kvs = self.ws.get(p)
        d = []
        for k in kvs:
            d.append(json.loads(k[1].get_value()))
        return d

    def get_network_router(self, sysid, tenantid, routerid):
        s = self.get_network_router_info_path(sysid, tenantid, routerid)
        kvs = self.ws.get(s)
        if len(kvs) == 0:
            return None
        return json.loads(kvs[0].get_value().get_value())

    def add_network_router(self, sysid, tenantid, routerid, routerinfo):
        p = self.get_network_router_info_path(sysid, tenantid, routerid)
        v = Value(json.dumps(routerinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_network_router(self, sysid, tenantid, routerid):
        p = self.get_network_router_info_path(sysid, tenantid, routerid)
        return self.ws.remove(p)

    def get_all_network_router(self, sysid, tenantid):
        p = self.get_all_routers_selector(sysid, tenantid)
        kvs = self.ws.get(p)
        d = []
        for k in kvs:
            d.append(json.loads(k[1].get_value()))
        return d

    def get_network(self, sysid, tenantid, netid):
        s = self.get_network_info_path(sysid, tenantid, netid)
        kvs = self.ws.get(s)
        if len(kvs) == 0:
            return None
        return json.loads(kvs[0].get_value().get_value())

    def get_all_networks (self, sysid, tenantid):
        p = self.get_all_networks_selector(sysid, tenantid)
        kvs = self.ws.get(p)
        d = []
        for k in kvs:
            d.append(json.loads(k.get_value().get_value()))
        return d

    def add_network(self, sysid, tenantid, netid, netinfo):
        p = self.get_network_info_path(sysid, tenantid, netid)
        v = Value(json.dumps(netinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_network(self, sysid, tenantid, netid):
        p = self.get_network_info_path(sysid, tenantid, netid)
        return self.ws.remove(p)

    # Images

    def add_image(self, sysid, tenantid, imageid, imginfo):
        p = self.get_image_info_path(sysid, tenantid, imageid)
        v = Value(json.dumps(imginfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_image(self, sysid, tenantid, imageid):
        p = self.get_image_info_path(sysid, tenantid, imageid)
        return self.ws.remove(p)

    def get_image(self, sysid, tenantid, imageid):
        p = self.get_image_info_path(sysid, tenantid, imageid)
        res = self.ws.get(p)
        if len(res) == 0:
            return None
        else:
            v = res[0].get_value()
            return json.loads(v.get_value())

    def get_all_images(self, sysid, tenantid):
        s = self.get_all_image_selector(sysid, tenantid)
        kvs = self.ws.get(s)
        d = []
        for n in kvs:
            d.append(json.loads(n.get_value().get_value()))
        return d

    # Node Images

    def add_node_image(self, sysid, tenantid, nodeid,imageid, imginfo):
        p = self.get_node_image_info_path(sysid, tenantid, nodeid, imageid)
        v = Value(json.dumps(imginfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_node_image(self, sysid, tenantid, nodeid,imageid):
        p = self.get_node_image_info_path(sysid, tenantid, nodeid, imageid)
        return self.ws.remove(p)

    def get_node_image(self, sysid, tenantid, nodeid,  imageid):
        p = self.get_node_image_info_path(sysid, tenantid, nodeid, imageid)
        res = self.ws.get(p)
        if len(res) == 0:
            return None
        else:
            v = res[0].get_value()
            return json.loads(v.get_value())

    def get_all_node_images(self, sysid, tenantid, nodeid):
        s = self.get_all_node_images(sysid, tenantid, nodeid)
        kvs = self.ws.get(s)
        d = []
        for n in kvs:
            d.append(json.loads(n[0][1].get_value()))
        return d


    # Flavor

    def add_flavor(self, sysid, tenantid, flavorid, flvinfo):
        p = self.get_flavor_info_path(sysid, tenantid, flavorid)
        v = Value(json.dumps(flvinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_flavor(self, sysid, tenantid, flavorid):
        p = self.get_flavor_info_path(sysid, tenantid, flavorid)
        return self.ws.remove(p)

    def get_flavor(self, sysid, tenantid, flavorid):
        p = self.get_flavor_info_path(sysid, tenantid, flavorid)
        res = self.ws.get(p)
        if len(res) == 0:
            return None
        else:
            v = res[0].get_value()
            return json.loads(v.get_value())

    def get_all_flavors(self, sysid, tenantid):
        s = self.get_all_flavor_selector(sysid, tenantid)
        kvs = self.ws.get(s)
        d = []
        for n in kvs:
            d.append(json.loads(n.get_value().get_value()))
        return d

    # Node Flavor

    def add_node_flavor(self, sysid, tenantid, nodeid,flavorid, flvinfo):
        p = self.get_node_flavor_info_path(sysid, tenantid, nodeid, flavorid)
        v = Value(json.dumps(flvinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_node_flavor(self, sysid, tenantid, nodeid,flavorid):
        p = self.get_node_flavor_info_path(sysid, tenantid, nodeid, flavorid)
        return self.ws.remove(p)

    def get_node_flavor(self, sysid, tenantid, nodeid,  flavorid):
        p = self.get_node_flavor_info_path(sysid, tenantid, nodeid, flavorid)
        res = self.ws.get(p)
        if len(res) == 0:
            return None
        else:
            v = res[0].get_value()
            return json.loads(v.get_value())

    def get_all_node_flavors(self, sysid, tenantid, nodeid):
        s = self.get_all_node_flavors(sysid, tenantid, nodeid)
        kvs = self.ws.get(s)
        d = []
        for n in kvs:
            d.append(json.loads(kvs[0].get_value().get_value()))
        return d

    # Node Network

    def add_node_network(self, sysid, tenantid, nodeid, netid, netinfo):
        p = self.get_node_network_info_path(sysid, tenantid, nodeid, netid)
        v = Value(json.dumps(netinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def get_node_network(self, sysid, tenantid, nodeid, netid):
        s = self.get_node_network_info_path(sysid, tenantid, nodeid, netid)
        kvs = self.ws.get(s)
        if len(kvs) == 0:
            return None
        return json.loads(kvs[0].get_value().get_value())

    def remove_node_network(self, sysid, tenantid, nodeid, netid):
        s = self.get_node_network_info_path(sysid, tenantid, nodeid, netid)
        return self.ws.remove(s)


    def add_node_floating_ip(self, sysid, tenantid, nodeid,floatingid, ip_info):

        p = self.get_node_network_floating_ip_info_path(sysid, tenantid, nodeid, floatingid)
        v = Value(json.dumps(ip_info), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_node_floating_ip(self, sysid, tenantid, nodeid, floatingid):
        p = self.get_node_network_floating_ip_info_path(sysid, tenantid, nodeid, floatingid)
        return self.ws.remove(p)

    def get_node_floating_ip(self, sysid, tenantid, nodeid,  floatingid):
        p = self.get_node_network_floating_ip_info_path(sysid, tenantid, nodeid, floatingid)
        res = self.ws.get(p)
        if len(res) == 0:
            return None
        else:
            v = res[0].get_value()
            return json.loads(v.get_value())

    def get_all_node_floating_ips(self, sysid, tenantid, nodeid):
        s = self.get_node_all_network_floating_ips_selector(sysid, tenantid, nodeid)
        kvs = self.ws.get(s)
        d = []
        for n in kvs:
            d.append(json.loads(kvs[0].get_value().get_value()))
        return d

    def get_all_nodes_network_ports(self, sysid, tenantid):
        s = self.get_node_network_ports_selector(sysid, tenantid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        xs = map(lambda x:
         (self.extract_nodeid_from_path(x.get_path()),
         self.extract_node_port_id_from_path(x.get_path()))
         ,res)
        return list(xs)

    def get_node_network_port(self, sysid, tenantid, nodeid,  portid):
        p = self.get_node_network_port_info_path(sysid, tenantid, nodeid, portid)
        res = self.ws.get(p)
        if len(res) == 0:
            return None
        else:
            v = res[0].get_value()
            return json.loads(v.get_value())

    def get_node_network_router(self, sysid, tenantid, nodeid, routerid):
        s = self.get_node_network_router_info_path(sysid, tenantid, nodeid, routerid)
        kvs = self.ws.get(s)
        if len(kvs) == 0:
            return None
        return json.loads(kvs[0].get_value().get_value())

    def add_node_network_router(self, sysid, tenantid, nodeid, routerid, routerinfo):
        p = self.get_node_network_router_info_path(sysid, tenantid, nodeid, routerid)
        v = Value(json.dumps(routerinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_node_network_router(self, sysid, tenantid, nodeid, routerid):
        p = self.get_node_network_router_info_path(sysid, tenantid, nodeid, routerid)
        return self.ws.remove(p)

    def get_all_node_network_routers(self, sysid, tenantid, nodeid):
        p = self.get_node_network_routers_selector(sysid, tenantid, nodeid)
        kvs = self.ws.get(p)
        d = []
        for k in kvs:
            d.append(json.loads(k[1].get_value()))
        return d

    def observe_node_routers(self, sysid, tenantid, nodeid, callback):
        s = self.get_node_network_routers_selector(sysid, tenantid, nodeid)

        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty datas')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    # Agent Evals

    def add_node_port_to_network(self, sysid, tenantid,  nodeid, portid, network_id):
        fname = 'add_port_to_network'
        params = {'cp_uuid': portid, 'network_uuid':network_id}
        s = self.get_agent_exec_path_with_params(sysid, tenantid, nodeid, fname, params)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def remove_node_port_from_network(self, sysid, tenantid, nodeid, portid):
        fname = 'remove_port_from_network'
        params = {'cp_uuid': portid}
        s = self.get_agent_exec_path_with_params(sysid, tenantid, nodeid, fname, params)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def add_node_floatingip(self, sysid, tenantid, nodeid):
        fname = 'create_floating_ip'
        s = self.get_agent_exec_path(sysid, tenantid, nodeid, fname)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def remove_node_floatingip(self, sysid, tenantid, nodeid, ipid):
        fname = 'delete_floating_ip'
        params = {'floating_uuid': ipid}
        s = self.get_agent_exec_path_with_params(sysid, tenantid, nodeid, fname, params)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def assign_node_floating_ip(self, sysid, tenantid, nodeid, ipid, cpid):
        fname = 'assign_floating_ip'
        params = {'floating_uuid': ipid, 'cp_uuid': cpid}
        s = self.get_agent_exec_path_with_params(sysid, tenantid, nodeid, fname, params)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def retain_node_floating_ip(self, sysid, tenantid, nodeid, ipid, cpid):
        fname = 'remove_floating_ip'
        params = {'floating_uuid': ipid, 'cp_uuid': cpid}
        s = self.get_agent_exec_path_with_params(sysid, tenantid, nodeid, fname, params)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def add_port_to_router(self, sysid, tenantid, nodeid, router_id, port_type, vnet_id=None, ip_address=None):
        fname = 'add_router_port'
        params = {'router_id': router_id, "port_type": port_type}
        if vnet_id is not None and vnet_id is not '':
            params.update({'vnet_id': vnet_id})
        if ip_address is not None and ip_address is not '':
            params.update({'ip_address': ip_address})
        s = self.get_agent_exec_path_with_params(sysid, tenantid, nodeid, fname, params)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def remove_port_from_router(self, sysid, tenantid, nodeid, router_id, vnet_id):
        fname = 'remove_router_port'
        params = {'router_id': router_id, "vnet_id": vnet_id}
        s = self.get_agent_exec_path_with_params(sysid, tenantid, nodeid, fname, params)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def onboard_fdu_from_node(self, sysid, tenantid, nodeid, fdu_id, fdu_info):
        fname = 'onboard_fdu'
        params = {'descriptor': fdu_info}
        s = self.get_agent_exec_path_with_params(sysid, tenantid, nodeid, fname, params)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def define_fdu_in_node(self, sysid, tenantid, nodeid, fdu_id):
        fname = 'define_fdu'
        params = {'fdu_id': fdu_id}
        s = self.get_agent_exec_path_with_params(sysid, tenantid, nodeid, fname, params)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def start_fdu_in_node(self, sysid, tenantid, instanceid, env):
        s = self.get_fdu_start_instance_selector(sysid, tenantid, instanceid, env)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on start_fdu_in_node')
        else:
            return json.loads(res[0].get_value().get_value())

    def run_fdu_in_node(self, sysid, tenantid, instanceid, env):
        s = self.get_fdu_run_instance_selector(sysid, tenantid, instanceid, env)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on run_fdu_in_node')
        else:
            return json.loads(res[0].get_value().get_value())

    def log_fdu_in_node(self, sysid, tenantid, instanceid):
        s = self.get_fdu_log_instance_selector(sysid, tenantid, instanceid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on log_fdu_in_node')
        else:
            return json.loads(res[0].get_value().get_value())

    def ls_fdu_in_node(self, sysid, tenantid, instanceid):
        s = self.get_fdu_ls_instance_selector(sysid, tenantid, instanceid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on ls_fdu_in_node')
        else:
            return json.loads(res[0].get_value().get_value())

    def file_fdu_in_node(self, sysid, tenantid, instanceid, filename):
        s = self.get_fdu_file_instance_selector(sysid, tenantid, instanceid, filename)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on file_fdu_in_node')
        else:
            return json.loads(res[0].get_value().get_value())

    # Scheduler
    def schedule_fdu_from_node(self, sysid, tenantid, nodeid, fduid):
        s = self.get_fdu_scheduler_selector(sysid, tenantid, nodeid, fduid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on file_fdu_in_node')
        else:
            return json.loads(res[0].get_value().get_value())


    def create_network_in_node(self, sysid, tenantid, nodeid, net_info):
        fname = 'create_node_network'
        params = {'descriptor': json.dumps(net_info)}
        s = self.get_agent_exec_path_with_params(sysid, tenantid, nodeid, fname, params)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def remove_network_from_node(self, sysid, tenantid, nodeid, net_id):
        fname = 'remove_node_network'
        params = {'net_id': net_id}
        s = self.get_agent_exec_path_with_params(sysid, tenantid, nodeid, fname, params)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())


class LAD(object):
    def __init__(self, workspace, prefix):
        self.ws = workspace
        self.prefix = prefix
        self.listeners = []
        self.evals = []

    def unsubscribe(self, subid):
        if subid in self.listeners:
            self.ws.unsubscribe(subid)
            self.listeners.remove(subid)

    def unregister_eval(self, path):
        if path in self.evals:
            self.ws.unregister_eval(path)
            self.evals.remove(path)

    def close(self):
        for s in self.listeners:
            self.ws.unsubscribe(s)
        for ep in self.evals:
            self.ws.unregister_eval(ep)

    # TODO: this should be in the YAKS api in the creation of a selector
    def dict2args(self, d):
        i = 0
        b = ''
        for k in d:
            v = d.get(k)
            if isinstance(v,(dict, list)):
                v = json.dumps(v)
            if i == 0:
                b = b + '{}={}'.format(k, v)
            else:
                b = b + ';{}={}'.format(k, v)
            i = i + 1
        return '('+b+')'

    # Node

    def get_node_info_path(self, nodeid):
        return Constants.create_path([self.prefix, nodeid, 'info'])

    def get_node_configuration_path(self, nodeid):
        return Constants.create_path([self.prefix, nodeid, 'configuration'])

    def get_node_status_path(self, nodeid):
        return Constants.create_path([self.prefix, nodeid, 'status'])

    def get_node_plugins_selector(self, nodeid):
        return Constants.create_path([self.prefix, nodeid,
                                      'plugins', '*', 'info'])

    def get_node_plugins_subscriber_selector(self, nodeid):
        return Constants.create_path([self.prefix, nodeid, 'plugins', '**'])

    def get_node_plugin_info_path(self, nodeid, pluginid):
        return Constants.create_path(
            [self.prefix, nodeid, 'plugins', pluginid, 'info'])

    def get_node_runtimes_selector(self, nodeid):
        return Constants.create_path([self.prefix, nodeid, 'runtimes', '**'])

    def get_node_network_managers_selector(self, nodeid):
        return Constants.create_path(
            [self.prefix, nodeid, 'network_managers', '*'])

    # Node FDU

    def get_node_runtime_fdus_selector(self, nodeid, pluginid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', pluginid,
                'fdu', '*','instances','*', 'info'])

    def get_node_runtime_fdus_subscriber_selector(self, nodeid, pluginid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', pluginid, 'fdu', '*',
            'instances','*','info'])

    def get_node_fdu_info_path(self, nodeid, pluginid, fduid, instanceid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', pluginid, 'fdu', fduid,
                'instances', instanceid, 'info'])

    def get_node_fdu_instances_selector(self, nodeid, fduid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', '*', 'fdu', fduid,
                'instances', '*', 'info'])

    def get_node_fdu_instance_selector(self, nodeid, instanceid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', '*', 'fdu', '*',
                'instances', instanceid, 'info'])

    def get_node_all_fdus_instances_selector(self, nodeid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', '*', 'fdu', '*',
            'instances','*','info'])

    def get_node_fdu_instance_start_selector(self, nodeid, instanceid, env):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', '*', 'fdu', '*',
                'instances', instanceid, 'start','?(env={})'.format(env)])

    def get_node_fdu_instance_run_selector(self, nodeid, instanceid, env):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', '*', 'fdu', '*',
                'instances', instanceid, 'run','?(env={})'.format(env)])

    def get_node_fdu_instance_log_selector(self, nodeid, instanceid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', '*', 'fdu', '*',
                'instances', instanceid, 'log'])

    def get_node_fdu_instance_ls_selector(self, nodeid, instanceid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', '*', 'fdu', '*',
                'instances', instanceid, 'ls'])

    def get_node_fdu_instance_file_selector(self, nodeid, instanceid, filename):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', '*', 'fdu', '*',
                'instances', instanceid, 'get', '?(filename={})'.format(filename)])

    def get_node_fdu_instance_start_path (self, nodeid, pluginid, fduid, instanceid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes',pluginid, 'fdu', fduid,
                'instances', instanceid, 'start'])

    def get_node_fdu_instance_run_path (self, nodeid, pluginid, fduid, instanceid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes',pluginid, 'fdu', fduid,
                'instances', instanceid, 'run'])

    def get_node_fdu_instance_log_path (self, nodeid, pluginid, fduid, instanceid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes',pluginid, 'fdu', fduid,
                'instances', instanceid, 'log'])

    def get_node_fdu_instance_ls_path (self, nodeid, pluginid, fduid, instanceid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes',pluginid, 'fdu', fduid,
                'instances', instanceid, 'ls'])

    def get_node_fdu_instance_file_path (self, nodeid, pluginid, fduid, instanceid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes',pluginid, 'fdu', fduid,
                'instances', instanceid, 'get'])

    # Node Images

    def get_node_image_info_path(self, nodeid, pluginid, imgid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', pluginid,
             'images', imgid, 'info'])

    # Node Flavors

    def get_node_flavor_info_path(self, nodeid, pluginid, flvid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', pluginid,
             'flavors', flvid, 'info'])

    # Node Networks

    def get_node_netwoks_selector(self, nodeid, pluginid):
        return Constants.create_path(
            [self.prefix, nodeid, 'network_managers',
             pluginid, 'networks', '*', 'info'])

    def get_node_netwoks_find_selector(self, nodeid, netid):
        return Constants.create_path(
            [self.prefix, nodeid, 'network_managers',
             '*', 'networks', netid, 'info'])

    def get_node_network_info_path(self, nodeid, pluginid, networkid):
        return Constants.create_path(
            [self.prefix, nodeid, 'network_managers',
             pluginid, 'networks', networkid, 'info'])

    def get_node_network_port_info_path(self, nodeid, pluginid,
                                        portid):
        return Constants.create_path(
            [self.prefix, nodeid, 'network_managers',
             pluginid, 'ports', portid, 'info'])

    def get_node_networks_port_selector(self, nodeid, pluginid):
        return Constants.create_path(
            [self.prefix, nodeid, 'network_managers',
             pluginid, 'ports', '*','info'])

    def get_node_network_router_info_path(self, nodeid, pluginid,
                                        routerid):
        return Constants.create_path(
            [self.prefix, nodeid, 'network_managers',
             pluginid, 'routers', routerid, 'info'])

    def get_node_network_routers_selector(self, nodeid, pluginid):
        return Constants.create_path(
            [self.prefix, nodeid, 'network_managers',
             pluginid, 'routers', '*','info'])

    def get_node_network_floating_ip_info_path(self, nodeid, pluginid, ipid):
        return Constants.create_path([self.prefix, nodeid,
            'network_managers',pluginid , 'floating-ips', ipid, 'info'])

    def get_node_all_network_floating_ips_selector(self, nodeid, pluginid):
        return Constants.create_path([self.prefix, nodeid,
            'network_managers', pluginid , 'floating-ips', '*', 'info'])

    # Node Evals

    def get_agent_exec_path(self, nodeid, func_name):
        return Constants.create_path([self.prefix, nodeid, 'agent', 'exec',
        func_name])

    def get_agent_exec_path_with_params(self, nodeid, func_name, params):
        if len(params) > 0:
            p = self.dict2args(params)
            f = func_name + '?' + p
        else:
            f = func_name
        return Constants.create_path([self.prefix, nodeid, 'agent', 'exec',
        f])

    def get_node_os_exec_path(self, nodeid, func_name):
        return Constants.create_path(
            [self.prefix, nodeid, 'os', 'exec', func_name])

    def get_node_nw_exec_path(self, nodeid, net_manager_uuid , func_name):
        return Constants.create_path(
            [self.prefix, nodeid, 'network_managers',
            net_manager_uuid , 'exec', func_name])

    def get_node_nw_exec_path_with_params(self, nodeid, net_manager_uuid,
        func_name, params):
        p = self.dict2args(params)
        f = func_name + '?' + p
        return Constants.create_path(
            [self.prefix, nodeid, 'network_managers',
            net_manager_uuid , 'exec', f])

    def get_node_os_exec_path_with_params(self, nodeid, func_name, params):
        if len(params) > 0:
            p = self.dict2args(params)
            f = func_name + '?' + p
        else:
            f = func_name
        return Constants.create_path(
            [self.prefix, nodeid, 'os', 'exec', f])

    def get_node_os_info_path(self, nodeid):
        return Constants.create_path(
            [self.prefix, nodeid, 'os', 'info'])

    def get_node_plugin_eval_path(self, nodeid, pluginid, func_name):
        return Constants.create_path(
            [self.prefix, nodeid, 'plugins', pluginid, 'exec', func_name])

    def get_node_plugin_eval_path_with_params(self, nodeid, pluginid,
                                              func_name, params):
        p = self.dict2args(params)
        f = func_name + '?' + p
        return Constants.create_path(
            [self.prefix, nodeid, 'plugins', pluginid,
             'exec', f])


    # ID Extraction

    def extract_nodeid_from_path(self, path):
        return path.split('/')[2]

    def extract_plugin_from_path(self, path):
        return path.split('/')[4]

    def extract_node_fduid_from_path(self, path):
        return path.split('/')[6]

    def extract_node_instanceid_from_path(self, path):
        return path.split('/')[8]

    def extract_node_routerid_from_path(self, path):
        return path.split('/')[6]


    # Node Evals

    def add_os_eval(self, nodeid, func_name, func):
        p = self.get_node_os_exec_path(nodeid, func_name)

        def cb(path, properties):
            try:
                v = Value(json.dumps(func(**properties)), encoding=Encoding.STRING)
                return v
            except Exception as e:
                return Value(json.dumps({'error':'{}'.format(e)}), encoding=Encoding.STRING)
        r = self.ws.register_eval(p, cb)
        self.evals.append(p)
        return r

    def exec_agent_eval(self, nodeid, func_name, parameters):
        s = self.get_agent_exec_path_with_params(
            nodeid, func_name, parameters)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def exec_os_eval(self, nodeid, func_name, parameters):
        s = self.get_node_os_exec_path_with_params(
            nodeid, func_name, parameters)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_os_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def add_nw_eval(self, nodeid, nm_uuid, func_name, func):
        p = self.get_node_nw_exec_path(nodeid, nm_uuid, func_name)

        def cb(path, properties):
            try:
                v = Value(json.dumps(func(**properties)), encoding=Encoding.STRING)
                return v
            except Exception as e:
                return Value(json.dumps({'error':'{}'.format(e)}), encoding=Encoding.STRING)
        r = self.ws.register_eval(p, cb)
        self.evals.append(p)
        return r

    def exec_nw_eval(self, nodeid, nm_uuid, func_name, parameters):
        s = self.get_node_nw_exec_path_with_params(
            nodeid, nm_uuid, func_name, parameters)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_nw_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def exec_plugin_eval(self, nodeid, pluginid, func_name, parameters):
        s = self.get_node_plugin_eval_path_with_params(
            nodeid, pluginid, func_name, parameters)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_plugin_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def add_plugin_eval(self, nodeid, pluginid, func_name, func):
        p = self.get_node_plugin_eval_path(nodeid, pluginid, func_name)

        def cb(path, properties):
            try:
                v = Value(json.dumps(func(**properties)), encoding=Encoding.STRING)
                return v
            except Exception as e:
                return Value(json.dumps({'error':'{}'.format(e)}), encoding=Encoding.STRING)
        r = self.ws.register_eval(p, cb)
        self.evals.append(p)
        return r

    def add_plugin_fdu_start_eval(self, nodeid, pluginid, fduid, instanceid, func):
        p = self.get_node_fdu_instance_start_path(nodeid, pluginid, fduid, instanceid)
        def cb(path, properties):
            try:
                v = Value(json.dumps(func(properties['env'])), encoding=Encoding.STRING)
                return v
            except Exception as e:
                return Value(json.dumps({'error':'{}'.format(e)}), encoding=Encoding.STRING)
        r = self.ws.register_eval(p, cb)
        self.evals.append(p)
        return r

    def add_plugin_fdu_run_eval(self, nodeid, pluginid, fduid, instanceid, func):
        p = self.get_node_fdu_instance_run_path(nodeid, pluginid, fduid, instanceid)
        def cb(path, properties):
            try:
                v = Value(json.dumps(func(properties['env'])), encoding=Encoding.STRING)
                return v
            except Exception as e:
                return Value(json.dumps({'error':'{}'.format(e)}), encoding=Encoding.STRING)
        r = self.ws.register_eval(p, cb)
        self.evals.append(p)
        return r

    def add_plugin_fdu_log_eval(self, nodeid, pluginid, fduid, instanceid, func):
        p = self.get_node_fdu_instance_log_path(nodeid, pluginid, fduid, instanceid)
        def cb(path, properties):
            v = Value(json.dumps(func(None)), encoding=Encoding.STRING)
            return v
        r = self.ws.register_eval(p, cb)
        self.evals.append(p)
        return r

    def add_plugin_fdu_ls_eval(self, nodeid, pluginid, fduid, instanceid, func):
        p = self.get_node_fdu_instance_ls_path(nodeid, pluginid, fduid, instanceid)
        def cb(path, properties):
            v = Value(json.dumps(func(None)), encoding=Encoding.STRING)
            return v
        r = self.ws.register_eval(p, cb)
        self.evals.append(p)
        return r

    def add_plugin_fdu_file_eval(self, nodeid, pluginid, fduid, instanceid, func):
        p = self.get_node_fdu_instance_file_path(nodeid, pluginid, fduid, instanceid)
        def cb(path, properties):
            try:
                v = Value(json.dumps(func(properties['filename'])), encoding=Encoding.STRING)
                return v
            except Exception as e:
                return Value(json.dumps({'error':'{}'.format(e)}), encoding=Encoding.STRING)
        r = self.ws.register_eval(p, cb)
        self.evals.append(p)
        return r

    def remove_plugin_fdu_start_eval(self, nodeid, pluginid, fduid, instanceid):
        p = self.get_node_fdu_instance_start_path(nodeid, pluginid, fduid, instanceid)
        r = self.ws.unregister_eval(p)
        self.evals.remove(p)
        return r

    def remove_plugin_fdu_run_eval(self, nodeid, pluginid, fduid, instanceid):
        p = self.get_node_fdu_instance_run_path(nodeid, pluginid, fduid, instanceid)
        r = self.ws.unregister_eval(p)
        self.evals.remove(p)
        return r

    def remove_plugin_fdu_log_eval(self, nodeid, pluginid, fduid, instanceid):
        p = self.get_node_fdu_instance_log_path(nodeid, pluginid, fduid, instanceid)
        r = self.ws.unregister_eval(p)
        self.evals.remove(p)
        return r

    def remove_plugin_fdu_ls_eval(self, nodeid, pluginid, fduid, instanceid):
        p = self.get_node_fdu_instance_ls_path(nodeid, pluginid, fduid, instanceid)
        r = self.ws.unregister_eval(p)
        self.evals.remove(p)
        return r

    def remove_plugin_fdu_file_eval(self, nodeid, pluginid, fduid, instanceid):
        p = self.get_node_fdu_instance_file_path(nodeid, pluginid, fduid, instanceid)
        r = self.ws.unregister_eval(p)
        self.evals.remove(p)
        return r


    # Node

    def add_node_plugin(self, nodeid, pluginid, plugininfo):
        p = self.get_node_plugin_info_path(nodeid, pluginid)
        v = Value(json.dumps(plugininfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def get_all_plugins(self, nodeid):
        s = self.get_node_plugins_selector(nodeid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        xs = map(lambda x: json.loads(x.get_value().get_value()), res)
        return list(xs)

    def add_node_information(self, nodeid, nodeinfo):
        p = self.get_node_info_path(nodeid)
        v = Value(json.dumps(nodeinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_node_information(self, nodeid):
        p = self.get_node_info_path(nodeid)
        self.ws.remove(p)

    def get_node_status(self, nodeid):
        p = self.get_node_status_path(nodeid)
        res = self.ws.get(p)
        if len(res) == 0:
            raise ValueError('Empty message list on get_node_status')
        v = res[0].get_value()
        return json.loads(v.get_value())

    def add_node_status(self, nodeid, nodestatus):
        p = self.get_node_status_path(nodeid)
        v = Value(json.dumps(nodestatus), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def observe_node_status(self, nodeid, callback):
        s = self.get_node_status_path(nodeid)
        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def remove_node_status(self, nodeid):
        p = self.get_node_status_path(nodeid)
        self.ws.remove(p)

    def get_node_configuration(self, nodeid):
        s = self.get_node_configuration_path(nodeid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on get_node_configuration')
        return json.loads(res[0].get_value().get_value())

    def observe_node_plugins(self, nodeid, callback):
        s = self.get_node_plugins_subscriber_selector(nodeid)

        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def observe_node_runtime_fdus(self, nodeid, pluginid, callback):
        s = self.get_node_runtime_fdus_subscriber_selector(nodeid, pluginid)

        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def get_node_info(self, nodeid):
        s = self.get_node_info_path(nodeid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on get_node_info')
        return json.loads(res[0].get_value().get_value())

    def get_node_os_info(self, nodeid):
        s = self.get_node_os_info_path(nodeid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on get_node_os_info')
        return json.loads(res[0].get_value().get_value())

    def add_node_os_info(self, nodeid, osinfo):
        p = self.get_node_os_info_path(nodeid)
        v = Value(json.dumps(osinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    # Node FDU

    def add_node_fdu(self, nodeid, pluginid, fduid, instanceid, fduinfo):
        p = self.get_node_fdu_info_path(nodeid, pluginid, fduid, instanceid)
        v = Value(json.dumps(fduinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def get_node_fdu(self, nodeid, pluginid, fduid, instanceid):
        s = self.get_node_fdu_info_path(nodeid, pluginid, fduid, instanceid)
        res = self.ws.get(s)
        if len(res) == 0:
            return None
        return json.loads(res[0].get_value().get_value())

    def remove_node_fdu(self, nodeid, pluginid, fduid, instanceid):
        p = self.get_node_fdu_info_path(nodeid, pluginid, fduid, instanceid)
        return self.ws.remove(p)

    def get_node_fdu_instances(self, nodeid, fduid):
        p = self.get_node_fdu_instances_selector(nodeid, fduid)
        res = self.ws.get(p)
        if len(res) == 0:
            return []
        return list(map(lambda x: self.extract_node_instanceid_from_path(x.get_path()), res))

    def get_node_all_fdus_instances(self, nodeid):
        p = self.get_node_all_fdus_instances_selector(nodeid)
        res = self.ws.get(p)
        if len(res) == 0:
            return []
        return list(map (lambda x: json.loads(x.get_value().get_value()), res))

    # Node Images

    def add_node_image(self, nodeid, pluginid, imgid, imginfo):
        p = self.get_node_image_info_path(nodeid, pluginid, imgid)
        v = Value(json.dumps(imginfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def get_node_image(self, nodeid, pluginid, imgid):
        s = self.get_node_image_info_path(nodeid, pluginid, imgid)
        res = self.ws.get(s)
        if len(res) == 0:
            return None
        return json.loads(res[0].get_value().get_value())

    def remove_node_image(self, nodeid, pluginid, imgid):
        p = self.get_node_image_info_path(nodeid, pluginid, imgid)
        return self.ws.remove(p)

    # Node Flavor

    def add_node_flavor(self, nodeid, pluginid, flvid, flvinfo):
        p = self.get_node_flavor_info_path(nodeid, pluginid, flvid)
        v = Value(json.dumps(flvinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def get_node_flavor(self, nodeid, pluginid, flvid):
        s = self.get_node_flavor_info_path(nodeid, pluginid, flvid)
        res = self.ws.get(s)
        if len(res) == 0:
            return None
        return json.loads(res[0].get_value().get_value())

    def remove_node_flavor(self, nodeid, pluginid, flvid):
        p = self.get_node_flavor_info_path(nodeid, pluginid, flvid)
        return self.ws.remove(p)

    #  Node Network

    def observe_node_networks(self, nodeid, pluginid, callback):
        s = self.get_node_netwoks_selector(nodeid, pluginid)

        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def add_node_network(self, nodeid, pluginid, netid, netinfo):
        p = self.get_node_network_info_path(nodeid, pluginid, netid)
        v = Value(json.dumps(netinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def get_node_network(self, nodeid, pluginid, netid):
        s = self.get_node_network_info_path(nodeid, pluginid, netid)
        res = self.ws.get(s)
        if len(res) == 0:
            return None
        return json.loads(res[0].get_value().get_value())

    def find_node_network(self, nodeid, netid):
        s = self.get_node_netwoks_find_selector(nodeid, netid)
        res = self.ws.get(s)
        if len(res) == 0:
            return None
        return json.loads(res[0].get_value().get_value())

    def remove_node_network(self, nodeid, pluginid, netid):
        p = self.get_node_network_info_path(nodeid, pluginid, netid)
        return self.ws.remove(p)

    def get_all_node_networks(self, nodeid, pluginid):
        s = self.get_node_netwoks_selector(nodeid, pluginid)
        kvs = self.ws.get(s)
        d = []
        for n in kvs:
            d.append(json.loads(kvs[0].get_value().get_value()))
        return d

    def add_node_port(self, nodeid, pluginid, portid, portinfo):
        p = self.get_node_network_port_info_path(nodeid, pluginid, portid)
        v = Value(json.dumps(portinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_node_port(self, nodeid, pluginid, portid):
        p = self.get_node_network_port_info_path(nodeid, pluginid, portid)
        return self.ws.remove(p)

    def get_node_port(self, nodeid, pluginid, portid):
        s = self.get_node_network_port_info_path(nodeid, pluginid, portid)
        res = self.ws.get(s)
        if len(res) == 0:
            return None
        return json.loads(res[0].get_value().get_value())

    def observe_node_ports(self, nodeid, pluginid, callback):
        s = self.get_node_networks_port_selector(nodeid, pluginid)

        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def get_all_node_ports(self, nodeid, pluginid):
        s = self.get_node_networks_port_selector(nodeid, pluginid)
        kvs = self.ws.get(s)
        d = []
        for n in kvs:
            d.append(json.loads(kvs[0].get_value().get_value()))
        return d

    def add_node_router(self, nodeid, pluginid, routerid, routerinfo):
        p = self.get_node_network_router_info_path(nodeid, pluginid, routerid)
        v = Value(json.dumps(routerinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_node_router(self, nodeid, pluginid, routerid):
        p = self.get_node_network_router_info_path(nodeid, pluginid, routerid)
        return self.ws.remove(p)

    def get_node_router(self, nodeid, pluginid, routerid):
        s = self.get_node_network_router_info_path(nodeid, pluginid, routerid)
        res = self.ws.get(s)
        if len(res) == 0:
            return None
        return json.loads(res[0].get_value().get_value())

    def observe_node_routers(self, nodeid, pluginid, callback):
        s = self.get_node_network_routers_selector(nodeid, pluginid)

        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty datas')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def get_all_node_routers(self, nodeid, pluginid):
        s = self.get_node_network_routers_selector(nodeid, pluginid)
        kvs = self.ws.get(s)
        d = []
        for n in kvs:
            d.append(json.loads(kvs[0].get_value().get_value()))
        return d

    def add_node_floating_ip(self, nodeid, pluginid, ipid, ipinfo):
        p = self.get_node_network_floating_ip_info_path(nodeid, pluginid, ipid)
        v = Value(json.dumps(ipinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def remove_node_floating_ip(self, nodeid, pluginid, ipid):
        p = self.get_node_network_floating_ip_info_path(nodeid, pluginid, ipid)
        return self.ws.remove(p)

    def get_node_floating_ip(self, nodeid, pluginid, ipid):
        s = self.get_node_network_floating_ip_info_path(nodeid, pluginid, ipid)
        res = self.ws.get(s)
        if len(res) == 0:
            return None
        return json.loads(res[0].get_value().get_value())

    def get_all_node_floating_ips(self, nodeid, pluginid):
        s = self.get_node_all_network_floating_ips_selector(nodeid, pluginid)
        kvs = self.ws.get(s)
        d = []
        for n in kvs:
            d.append(json.loads(kvs[0].get_value().get_value()))
        return d

    def observe_node_floating_ip(self, nodeid, pluginid, callback):
        s = self.get_node_all_network_floating_ips_selector(nodeid, pluginid)
        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty datas')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid


class CLAD(object):
    def __init__(self, workspace, prefix):
        self.ws = workspace
        self.prefix = prefix
        self.listeners = []
        self.evals = []

    def unsubscribe(self, subid):
        if subid in self.listeners:
            self.ws.unsubscribe(subid)
            self.listeners.remove(subid)

    def unregister_eval(self, path):
        if path in self.evals:
            self.ws.unregister_eval(path)
            self.evals.remove(path)

    def close(self):
        for s in self.listeners:
            self.ws.unsubscribe(s)
        for ep in self.evals:
            self.ws.unregister_eval(ep)

    # TODO: this should be in the YAKS api in the creation of a selector
    def dict2args(self, d):
        i = 0
        b = ''
        for k in d:
            v = d.get(k)
            if isinstance(v,(dict, list)):
                v = json.dumps(v)
            if i == 0:
                b = b + '{}={}'.format(k, v)
            else:
                b = b + ';{}={}'.format(k, v)
            i = i + 1
        return '('+b+')'

    # Node

    def get_node_selector(self):
        return Constants.create_path([self.prefix, '*', 'info'])

    def get_node_info_path(self, nodeid):
        return Constants.create_path([self.prefix, nodeid, 'info'])

    def get_node_status_path(self, nodeid):
        return Constants.create_path([self.prefix, nodeid, 'status'])

    def get_node_configuration_path(self, nodeid):
        return Constants.create_path([self.prefix, nodeid, 'configuration'])

    def get_node_plugins_selector(self, nodeid):
        return Constants.create_path([self.prefix, nodeid,
                                      'plugins', '*', 'info'])

    def get_node_plugins_subscriber_selector(self, nodeid):
        return Constants.create_path([self.prefix, nodeid, 'plugins', '**'])

    def get_node_plugin_info_path(self, nodeid, pluginid):
        return Constants.create_path(
            [self.prefix, nodeid, 'plugins', pluginid, 'info'])

    def get_node_runtimes_selector(self, nodeid):
        return Constants.create_path([self.prefix, nodeid, 'runtimes', '**'])

    def get_node_runtime_fdus_selector(self, nodeid, pluginid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', pluginid, 'fdu', '*', 'info'])

    def get_node_runtime_fdus_subscriber_selector(self, nodeid, pluginid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', pluginid, 'fdu', '**'])

    def get_node_status(self, nodeid):
        p = self.get_node_status_path(nodeid)
        res = self.ws.get(p)
        if len(res) == 0:
            raise ValueError('Empty message list on get_node_status')
        else:
            v = res[0].get_value()
            return json.loads(v.get_value())

    def add_node_status(self, nodeid, nodestatus):
        p = self.get_node_status_path(nodeid)
        v = Value(json.dumps(nodestatus), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def observe_node_status(self, nodeid, callback):
        s = self.get_node_status_path(nodeid)
        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    def remove_node_status(self, nodeid):
        p = self.get_node_status_path(nodeid)
        self.ws.remove(p)

    # Node FDU

    def get_node_fdu_info_path(self, nodeid, pluginid, fduid):
        return Constants.create_path(
            [self.prefix, nodeid, 'runtimes', pluginid, 'fdu', fduid, 'info'])

    # Agent Evals

    def get_agent_exec_path(self, nodeid, func_name):
        return Constants.create_path([self.prefix, nodeid, 'agent', 'exec',
        func_name])

    def get_agent_exec_path_with_params(self, nodeid, func_name, params):
        p = self.dict2args(params)
        f = func_name + '?' + p
        return Constants.create_path([self.prefix, nodeid, 'agent', 'exec',
        f])

    def exec_agent_eval(self, nodeid, func_name, parameters):
        s = self.get_agent_exec_path_with_params(
            nodeid, func_name, parameters)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_agent_eval')
        else:
            return json.loads(res[0].get_value().get_value())


    # Node Evals

    def get_node_os_exec_path(self, nodeid, func_name):
        return Constants.create_path(
            [self.prefix, nodeid, 'os', 'exec', func_name])

    def get_node_os_exec_path_with_params(self, nodeid, func_name, params):
        p = self.dict2args(params)
        f = func_name + '?' + p
        return Constants.create_path(
            [self.prefix, nodeid, 'os', 'exec', f])

    def get_node_os_info_path(self, nodeid):
        return Constants.create_path(
            [self.prefix, nodeid, 'os', 'info'])

    def get_node_plugin_eval_path(self, nodeid, pluginid, func_name):
        return Constants.create_path(
            [self.prefix, nodeid, 'plugins', pluginid, 'exec', func_name])

    def get_node_plugin_eval_path_with_params(self, nodeid, pluginid,
                                              func_name, params):
        p = self.dict2args(params)
        f = func_name + '?' + p
        return Constants.create_path(
            [self.prefix, nodeid, 'plugins', pluginid,
             'exec', f])

    def add_os_eval(self, nodeid, func_name, func):
        p = self.get_node_os_exec_path(nodeid, func_name)

        def cb(path, properties):
            v = Value(json.dumps(func(**properties)), encoding=Encoding.STRING)
            return v
        r = self.ws.register_eval(p, cb)
        self.evals.append(p)
        return r

    def exec_os_eval(self, nodeid, func_name, parameters):
        s = self.get_node_os_exec_path_with_params(
            nodeid, func_name, parameters)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_os_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def exec_plugin_eval(self, nodeid, pluginid, func_name, parameters):
        s = self.get_node_plugin_eval_path_with_params(
            nodeid, pluginid, func_name, parameters)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on exec_os_eval')
        else:
            return json.loads(res[0].get_value().get_value())

    def add_plugin_eval(self, nodeid, pluginid, func_name, func):
        p = self.get_node_plugin_eval_path(nodeid, pluginid, func_name)

        def cb(path, properties):
            v = Value(json.dumps(func(**properties)), encoding=Encoding.STRING)
            return v
        r = self.ws.register_eval(p, cb)
        self.evals.append(p)
        return r

    def add_node_plugin(self, nodeid, pluginid, plugininfo):
        p = self.get_node_plugin_info_path(nodeid, pluginid)
        v = Value(json.dumps(plugininfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def get_all_plugins(self, nodeid):
        s = self.get_node_plugins_selector(nodeid)
        res = self.ws.get(s)
        if len(res) == 0:
            return []
        xs = map(lambda x: json.loads(x.get_value().get_value()), res)
        return list(xs)

    def add_node_information(self, nodeid, nodeinfo):
        p = self.get_node_info_path(nodeid)
        v = Value(json.dumps(nodeinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def get_node_configuration(self, nodeid):
        s = self.get_node_configuration_path(nodeid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on get_node_configuration')
        return json.loads(res[0].get_value().get_value())

    def observe_node_plugins(self, nodeid, callback):
        s = self.get_node_plugins_subscriber_selector(nodeid)

        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    # Node FDU

    def add_node_fdu(self, nodeid, pluginid, fduid, fduinfo):
        p = self.get_node_fdu_info_path(nodeid, pluginid, fduid)
        v = Value(json.dumps(fduinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)

    def get_node_fdu(self, nodeid, pluginid, fduid):
        s = self.get_node_fdu_info_path(nodeid, pluginid, fduid)
        res = self.ws.get(s)
        if len(res) == 0:
            return None
        return json.loads(res[0].get_value().get_value())

    def remove_node_fdu(self, nodeid, pluginid, fduid):
        p = self.get_node_fdu_info_path(nodeid, pluginid, fduid)
        return self.ws.remove(p)

    def observe_node_runtime_fdus(self, nodeid, pluginid, callback):
        s = self.get_node_runtime_fdus_subscriber_selector(nodeid, pluginid)

        def cb(kvs):
            if len(kvs) == 0:
                raise ValueError('Listener received empty data')
            else:
                v = kvs[0].get_value()
                if v is not None:
                    callback(json.loads(v.get_value()))
        subid = self.ws.subscribe(s, cb)
        self.listeners.append(subid)
        return subid

    # Node

    def get_node_info(self, nodeid):
        s = self.get_node_info_path(nodeid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on get_node_info')
        return json.loads(res[0].get_value().get_value())

    def get_node_os_info(self, nodeid):
        s = self.get_node_os_info_path(nodeid)
        res = self.ws.get(s)
        if len(res) == 0:
            raise ValueError('Empty data on get_node_os_info')
        return json.loads(res[0].get_value().get_value())

    def add_node_os_info(self, nodeid, osinfo):
        p = self.get_node_os_info_path(nodeid)
        v = Value(json.dumps(osinfo), encoding=Encoding.STRING)
        return self.ws.put(p, v)


class Global(object):
    def __init__(self, workspace):
        self.ws = workspace
        self.actual = GAD(workspace, Constants.global_actual_prefix)
        self.desired = GAD(workspace, Constants.global_desired_prefix)

    def close(self):
        self.actual.close()
        self.desired.close()


class Local(object):
    def __init__(self, workspace):
        self.ws = workspace
        self.actual = LAD(workspace, Constants.local_actual_prefix)
        self.desired = LAD(workspace, Constants.local_desired_prefix)

    def close(self):
        self.actual.close()
        self.desired.close()



class Yaks_Connector(object):
    '''
    Class: Yaks_Connector

    This is an utility class to create Path, Selector and wraps YAKS API
    Users are not supposed to directly use methods from this class and its components
    '''
    def __init__(self, locator):
        self.executor = concurrent.futures.ThreadPoolExecutor()
        self.yaks_client = Yaks.login(locator)
        self.yaks_admin = self.yaks_client.admin()
        self.ws = self.yaks_client.workspace(Constants.global_actual_prefix, self.executor)
        self.glob = Global(self.ws)
        self.loc = Local(self.ws)

    def close(self):
        self.glob.close()
        self.loc.close()
        self.yaks_client.logout()

class Yaks_Constraint_Connector(object):
    def __init__(self, locator):
        self.executor = concurrent.futures.ThreadPoolExecutor()
        self.yaks_client = Yaks.login(locator)
        self.yaks_admin = self.yaks_client.admin()
        self.ws = self.yaks_client.workspace(Constants.local_constraint_actual_prefix, self.executor)
        self.actual = CLAD(self.ws, Constants.local_constraint_actual_prefix)
        self.desired = CLAD(self.ws, Constants.local_constaint_desired_prefix)

    def close(self):
        self.actual.close()
        self.desired.close()
        self.yaks_client.logout()



# Copyright (c) 2014,2018 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc.
# OCaml implementation and API

#!/usr/bin/env python3

from setuptools import setup

setup(
    name='fog05-sdk',
    version='0.2.0',
    python_requires='>=3',
    author='ADLINK',
    packages=['fog05_sdk','fog05_sdk.interfaces', 'fog05_sdk.types'],
    description='Eclipse fog05 Python 3 SDK',
    url='https://fog05.io',
    author_email='gabriele.baldoni@adlinktech.com',
    license='Apache 2.O or EPL 2.0',
    install_requires=['yaks==0.3.0.post1', 'jsonschema','mvar==0.0.1', 'pyangbind==0.8.1'],
    scripts=[],
    include_package_data=True
)



# Copyright (c) 2014,2018 Contributors to the Eclipse Foundation
#
# See the NOTICE file(s) distributed with this work for additional
# information regarding copyright ownership.
#
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0, or the Apache License, Version 2.0
# which is available at https://www.apache.org/licenses/LICENSE-2.0.
#
# SPDX-License-Identifier: EPL-2.0 OR Apache-2.0
#
# Contributors: Gabriele Baldoni, ADLINK Technology Inc.
# Initial implementation and API


from fog05_sdk.interfaces.Plugin import Plugin


class MonitoringPlugin(Plugin):

    def __init__(self, version, plugin_uuid=None):
        super(MonitoringPlugin, self).__init__(version, plugin_uuid)
        self.name = ''

    def start_monitoring(self):
        '''
        start the runtime
        :return: runtime pid or runtime uuid?
        '''
        raise NotImplementedError('This is and interface!')

    def stop_monitoring(self):
        '''
        stop this runtime
        '''
        raise NotImplementedError('This is and interface!')
